from pypresence import Client
import time

client_id = "743296148592263240"
RPC = Client(client_id)
RPC.start()


def register_join_event(syncplay_join_group: callable):
    RPC.register_event("activity_join", syncplay_join_group)


def send_presence(
    title: str,
    subtitle: str,
    playback_time: float = None,
    duration: float = None,
    playing: bool = False,
    syncplay_group: str = None,
):
    small_image = "play-dark3" if playing else None
    start = None
    end = None
    if playback_time is not None and duration is not None and playing:
        start = int(time.time() - playback_time)
        end = int(start + duration)

    payload = {
        "state": subtitle if subtitle else "Unknown Media",
        "details": title,
        "instance": False,
        "large_image": "jellyfin2",
        "start": start,
        "end": end,
        "large_text": "Jellyfin",
        "small_image": small_image,
    }

    if syncplay_group:
        payload["party_id"] = str(hash(syncplay_group))
        payload["party_size"] = [1, 100]
        payload["join"] = syncplay_group

    RPC.set_activity(**payload)


def clear_presence():
    RPC.clear_activity()



from .conf import settings
from .i18n import _

import urllib.request
import urllib.error
import logging
import sys
import time

log = logging.getLogger("svp_integration")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .menu import OSDMenu as OSDMenu_type
    from .player import PlayerManager as PlayerManager_type


def list_request(path: str):
    try:
        response = urllib.request.urlopen(settings.svp_url + "?" + path)
        return response.read().decode("utf-8").replace("\r\n", "\n").split("\n")
    except urllib.error.URLError:
        log.error("Could not reach SVP API server.", exc_info=True)
        return None


def simple_request(path: str):
    response_list = list_request(path)
    if response_list is None:
        return None
    if len(response_list) != 1 or " = " not in response_list[0]:
        return None
    return response_list[0].split(" = ")[1]


def get_profiles():
    profile_ids = list_request("list=profiles")
    profiles = {}
    for profile_id in profile_ids:
        profile_id = profile_id.replace("profiles.", "")
        if profile_id == "predef":
            continue
        if profile_id == "P10000001_1001_1001_1001_100000000001":
            profile_name = _("Automatic")
        else:
            profile_name = simple_request("profiles.{0}.title".format(profile_id))
        if simple_request("profiles.{0}.on".format(profile_id)) == "false":
            continue
        profile_guid = "{" + profile_id[1:].replace("_", "-") + "}"
        profiles[profile_guid] = profile_name
    return profiles


def get_name_from_guid(profile_id: str):
    profile_id = "P" + profile_id[1:-1].replace("-", "_")
    if profile_id == "P10000001_1001_1001_1001_100000000001":
        return _("Automatic")
    else:
        return simple_request("profiles.{0}.title".format(profile_id))


def get_last_profile():
    return simple_request("rt.playback.last_profile")


def is_svp_alive():
    try:
        response = list_request("")
        return response is not None
    except Exception:
        log.error("Could not reach SVP API server.", exc_info=True)
        return False


def is_svp_enabled():
    return simple_request("rt.disabled") == "false"


def is_svp_active():
    response = simple_request("rt.playback.active")
    if response is None:
        return False
    return response != ""


def set_active_profile(profile_id: str):
    # As far as I know, there is no way to directly set the profile.
    if not is_svp_active():
        return False
    if profile_id == get_last_profile():
        return True
    for i in range(len(list_request("list=profiles"))):
        list_request("!profile_next")
        if get_last_profile() == profile_id:
            return True
    return False


def set_disabled(disabled: bool):
    return (
        simple_request("rt.disabled={0}".format("true" if disabled else "false"))
        == "true"
    )


class SVPManager:
    def __init__(self, menu: "OSDMenu_type", player_manager: "PlayerManager_type"):
        self.menu = menu

        if settings.svp_enable:
            socket = settings.svp_socket
            if socket is None:
                if sys.platform.startswith("win32") or sys.platform.startswith(
                    "cygwin"
                ):
                    socket = "mpvpipe"
                else:
                    socket = "/tmp/mpvsocket"

            # This actually *adds* another ipc server.
            player_manager.add_ipc(socket)

        if settings.svp_enable and not is_svp_alive():
            log.error(
                "SVP is not reachable. Please make sure you have the API enabled."
            )

    @staticmethod
    def is_available():
        if not settings.svp_enable:
            return False
        if not is_svp_alive():
            return False
        return True

    def menu_set_profile(self):
        profile_id = self.menu.menu_list[self.menu.menu_selection][2]
        if profile_id is None:
            set_disabled(True)
        else:
            set_active_profile(profile_id)
        # Need to re-render menu. SVP has a race condition so we wait a second.
        time.sleep(1)
        self.menu.menu_action("back")
        self.menu_action()

    def menu_set_enabled(self):
        set_disabled(False)

        # Need to re-render menu. SVP has a race condition so we wait a second.
        time.sleep(1)
        self.menu.menu_action("back")
        self.menu_action()

    def menu_action(self):
        if is_svp_active():
            selected = 0
            active_profile = get_last_profile()
            profile_option_list = [(_("Disabled"), self.menu_set_profile, None)]
            for i, (profile_id, profile_name) in enumerate(get_profiles().items()):
                profile_option_list.append(
                    (profile_name, self.menu_set_profile, profile_id)
                )
                if profile_id == active_profile:
                    selected = i + 1
            self.menu.put_menu(_("Select SVP Profile"), profile_option_list, selected)
        else:
            if is_svp_enabled():
                self.menu.put_menu(
                    _("SVP is Not Active"),
                    [
                        (_("Disable"), self.menu_set_profile, None),
                        (_("Retry"), self.menu_set_enabled),
                    ],
                    selected=1,
                )
            else:
                self.menu.put_menu(
                    _("SVP is Disabled"), [(_("Enable SVP"), self.menu_set_enabled)]
                )



from __future__ import annotations
from typing import Optional
import logging

log = logging.getLogger("settings_base")

# This is NOT a full pydantic replacement!!!
# Compatible with PEP 563
# Tries to also deal with common errors in the config


def allow_none(constructor):
    def wrapper(input):
        if input is None or input == "null":
            return None
        return constructor(input)

    return wrapper


yes_set = {1, "yes", "Yes", "True", "true", "1", True}


def adv_bool(value):
    return value in yes_set


object_types = {
    "float": float,
    "int": int,
    "str": str,
    "bool": adv_bool,
    "Optional[float]": allow_none(float),
    "Optional[int]": allow_none(int),
    "Optional[str]": allow_none(str),
    "Optional[bool]": allow_none(adv_bool),
    float: float,
    int: int,
    str: str,
    bool: adv_bool,
    Optional[float]: allow_none(float),
    Optional[int]: allow_none(int),
    Optional[str]: allow_none(str),
    Optional[bool]: allow_none(adv_bool),
}


class SettingsBase:
    def __init__(self):
        self.__fields_set__ = set()
        self.__fields__ = []
        for attr in self.__class__.__annotations__.keys():
            if attr.startswith("_"):
                continue

            self.__fields__.append(attr)
            setattr(self, attr, getattr(self.__class__, attr))

    def dict(self):
        result = {}
        for attr in self.__fields__:
            result[attr] = getattr(self, attr)
        return result

    def parse_obj(self, object):
        new_obj = self.__class__()
        annotations = self.__class__.__annotations__

        for attr in self.__fields__:
            if attr in object:
                parse = object_types[annotations[attr]]
                try:
                    setattr(new_obj, attr, parse(object[attr]))
                    new_obj.__fields_set__.add(attr)
                except:
                    log.error(
                        "Setting {0} had invalid value {1}.".format(attr, object[attr]),
                        exc_info=True,
                    )
        return new_obj



import socket
import ipaddress
import requests
import urllib.parse
from threading import Lock
import logging
import sys
import os.path
import platform

from .conf import settings
from datetime import datetime
from functools import wraps
from .constants import USER_APP_NAME
from .i18n import _

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from jellyfin_apiclient_python import JellyfinClient as JellyfinClient_type

log = logging.getLogger("utils")

seq_num = 0
seq_num_lock = Lock()


class Timer(object):
    def __init__(self):
        self.started = datetime.now()

    def restart(self):
        self.started = datetime.now()

    def elapsed_ms(self):
        return self.elapsed() * 1e3

    def elapsed(self):
        return (datetime.now() - self.started).total_seconds()


def synchronous(tlockname: str):
    """
    A decorator to place an instance based lock around a method.
    From: http://code.activestate.com/recipes/577105-synchronization-decorator-for-class-methods/
    """

    def _synched(func):
        @wraps(func)
        def _synchronizer(self, *args, **kwargs):
            tlock = self.__getattribute__(tlockname)
            tlock.acquire()
            try:
                return func(self, *args, **kwargs)
            finally:
                tlock.release()

        return _synchronizer

    return _synched


def is_local_domain(client: "JellyfinClient_type"):
    # With Jellyfin, it is significantly more likely the user will be using
    # an address that is a hairpin NAT. We want to detect this and avoid
    # imposing limits in this case.
    url = client.config.data.get("auth.server", "")
    domain = urllib.parse.urlparse(url).hostname

    addr_info = socket.getaddrinfo(domain, 8096)[0]
    ip = addr_info[4][0]
    is_local = ipaddress.ip_address(ip).is_private

    if not is_local:
        if addr_info[0] == socket.AddressFamily.AF_INET:
            try:
                wan_ip = requests.get(
                    "https://checkip.amazonaws.com/", timeout=(3, 10)
                ).text.strip("\r\n")
                return ip == wan_ip
            except Exception:
                log.warning(
                    "checkip.amazonaws.com is unavailable. Assuming potential WAN ip is remote.",
                    exc_info=True,
                )
                return False
        elif addr_info[0] == socket.AddressFamily.AF_INET6:
            return False
    return True


def mpv_color_to_plex(color: str):
    return "#" + color.lower()[3:]


def plex_color_to_mpv(color: str):
    return "#FF" + color.upper()[1:]


def get_profile(
    is_remote: bool = False,
    video_bitrate: Optional[int] = None,
    force_transcode: bool = False,
    is_tv: bool = False,
):
    if video_bitrate is None:
        if is_remote:
            video_bitrate = settings.remote_kbps
        else:
            video_bitrate = settings.local_kbps

    if settings.force_video_codec:
        transcode_codecs = settings.force_video_codec
    elif settings.allow_transcode_to_h265 and not settings.transcode_hevc:
        transcode_codecs = "h264,h265,hevc,mpeg4,mpeg2video"
    elif settings.prefer_transcode_to_h265 and not settings.transcode_hevc:
        transcode_codecs = "h265,hevc,h264,mpeg4,mpeg2video"
    else:
        transcode_codecs = "h264,mpeg4,mpeg2video"

    if settings.force_audio_codec:
        audio_transcode_codecs = settings.force_audio_codec
    else:
        audio_transcode_codecs = "aac,mp3,ac3,opus,flac,vorbis"

    profile = {
        "Name": USER_APP_NAME,
        "MaxStreamingBitrate": video_bitrate * 1000,
        "MaxStaticBitrate": video_bitrate * 1000,
        "MusicStreamingTranscodingBitrate": 1280000,
        "TimelineOffsetSeconds": 5,
        "TranscodingProfiles": [
            {"Type": "Audio"},
            {
                "Container": "ts",
                "Type": "Video",
                "Protocol": "hls",
                "AudioCodec": audio_transcode_codecs,
                "VideoCodec": transcode_codecs,
                "MaxAudioChannels": "6",
            },
            {"Container": "jpeg", "Type": "Photo"},
        ],
        "DirectPlayProfiles": [{"Type": "Video"}, {"Type": "Audio"}, {"Type": "Photo"}],
        "ResponseProfiles": [],
        "ContainerProfiles": [],
        "CodecProfiles": [],
        "SubtitleProfiles": [
            {"Format": "srt", "Method": "External"},
            {"Format": "srt", "Method": "Embed"},
            {"Format": "ass", "Method": "External"},
            {"Format": "ass", "Method": "Embed"},
            {"Format": "sub", "Method": "Embed"},
            {"Format": "sub", "Method": "External"},
            {"Format": "ssa", "Method": "Embed"},
            {"Format": "ssa", "Method": "External"},
            {"Format": "smi", "Method": "Embed"},
            {"Format": "smi", "Method": "External"},
            # Jellyfin currently refuses to serve these subtitle types as external.
            {"Format": "pgssub", "Method": "Embed"},
            # {
            #    "Format": "pgssub",
            #    "Method": "External"
            # },
            {"Format": "dvdsub", "Method": "Embed"},
            {"Format": "dvbsub", "Method": "Embed"},
            # {
            #    "Format": "dvdsub",
            #    "Method": "External"
            # },
            {"Format": "pgs", "Method": "Embed"},
            # {
            #    "Format": "pgs",
            #    "Method": "External"
            # }
        ],
    }

    if settings.transcode_hi10p:
        profile["CodecProfiles"].append(
            {
                "Type": "Video",
                "Conditions": [
                    {
                        "Condition": "LessThanEqual",
                        "Property": "VideoBitDepth",
                        "Value": "8",
                    }
                ],
            }
        )

    if settings.transcode_dolby_vision:
        profile["CodecProfiles"].append(
            {
                "Type": "Video",
                "Conditions": [
                    {
                        "Condition": "NotEquals",
                        "Property": "VideoRangeType",
                        "Value": "DOVI",
                    }
                ],
            }
        )

    if settings.transcode_hdr:
        profile["CodecProfiles"].append(
            {
                "Type": "Video",
                "Conditions": [
                    {
                        "Condition": "Equals",
                        "Property": "VideoRangeType",
                        "Value": "SDR",
                    }
                ],
            }
        )

    if settings.transcode_hevc:
        profile["CodecProfiles"].append(
            {
                "Type": "Video",
                "Codec": "hevc",
                "Conditions": [
                    {
                        "Condition": "Equals",
                        "Property": "Width",
                        "Value": "0",
                    }
                ],
            }
        )
        profile["CodecProfiles"].append(
            {
                "Type": "Video",
                "Codec": "h265",
                "Conditions": [
                    {
                        "Condition": "Equals",
                        "Property": "Width",
                        "Value": "0",
                    }
                ],
            }
        )

    if settings.transcode_av1:
        profile["CodecProfiles"].append(
            {
                "Type": "Video",
                "Codec": "av1",
                "Conditions": [
                    {
                        "Condition": "Equals",
                        "Property": "Width",
                        "Value": "0",
                    }
                ],
            }
        )

    if settings.transcode_4k:
        profile["CodecProfiles"].append(
            {
                "Type": "Video",
                "Conditions": [
                    {
                        "Condition": "LessThanEqual",
                        "Property": "Width",
                        "Value": "1920",
                    },
                    {
                        "Condition": "LessThanEqual",
                        "Property": "Height",
                        "Value": "1080",
                    },
                ],
            }
        )

    if settings.always_transcode or force_transcode:
        profile["DirectPlayProfiles"] = []

    if is_tv:
        profile["TranscodingProfiles"].insert(
            0,
            {
                "Container": "ts",
                "Type": "Video",
                "AudioCodec": "mp3,aac",
                "VideoCodec": "h264",
                "Context": "Streaming",
                "Protocol": "hls",
                "MaxAudioChannels": "2",
                "MinSegments": "1",
                "BreakOnNonKeyFrames": True,
            },
        )

    return profile


def get_sub_display_title(stream: dict):
    return "{0}{1} ({2})".format(
        stream.get("Language", _("Unkn")).capitalize(),
        _(" Forced") if stream.get("IsForced") else "",
        stream.get("Codec"),
    )


def get_seq():
    global seq_num
    seq_num_lock.acquire()
    current = seq_num
    seq_num += 1
    seq_num_lock.release()
    return current


def none_fallback(value, fallback):
    if value is None:
        return fallback
    return value


def get_resource(*path):
    # Detect if bundled via pyinstaller.
    # From: https://stackoverflow.com/questions/404744/
    if getattr(sys, "_MEIPASS", False):
        application_path = os.path.join(getattr(sys, "_MEIPASS"), "jellyfin_mpv_shim")
    else:
        application_path = os.path.dirname(os.path.abspath(__file__))

    # ! Test code for Mac
    if getattr(sys, "frozen", False) and platform.system() == "Darwin":
        application_path = os.path.join(os.path.dirname(sys.executable), "../Resources")

    return os.path.join(application_path, *path)


def get_text(*path):
    with open(get_resource(*path)) as fh:
        return fh.read()



import logging
import threading
import os
from datetime import datetime, timedelta

from .clients import clientManager
from .media import Media
from .i18n import _

# This is based on: https://github.com/jellyfin/jellyfin-web/blob/master/src/components/syncPlay/syncPlayManager.js

from .conf import settings
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .player import PlayerManager as PlayerManager_type

log = logging.getLogger("syncplay")
seconds_in_ticks = 10000000
info_commands = {
    "GroupDoesNotExist": _("The specified SyncPlay group does not exist."),
    "CreateGroupDenied": _("Creating SyncPlay groups is not allowed."),
    "JoinGroupDenied": _("SyncPlay group access was denied."),
    "LibraryAccessDenied": _("Access to the SyncPlay library was denied."),
}


def _parse_precise_time(time: str):
    # We have to remove the Z and the least significant digit.
    return datetime.strptime(time[:-2], "%Y-%m-%dT%H:%M:%S.%f")


class TimeoutThread(threading.Thread):
    def __init__(self, action, delay: float, args):
        self.action = action
        self.delay = delay
        self.args = args
        self.halt = threading.Event()
        threading.Thread.__init__(self)
        self.daemon = True

    def run(self):
        if not self.halt.wait(timeout=self.delay / 1000):
            try:
                self.action(*self.args)
            except:
                log.error("TimeoutThread crashed.", exc_info=True)

    def stop(self):
        self.halt.set()
        self.join()


def set_timeout(ms: float, callback, *args):
    """Similar to setTimeout JS function."""
    timeout = TimeoutThread(callback, ms, args)
    timeout.start()
    return timeout.stop


class SyncPlayManager:
    def __init__(self, manager: "PlayerManager_type"):
        self.playerManager = manager
        self.menu = manager.menu

        self.sync_enabled = False
        self.playback_diff_ms = 0
        self.method = None
        self.attempts = 0
        self.last_sync_time = datetime.utcnow()
        self.enable_speed_sync = True

        self.last_playback_waiting = None
        self.min_buffer_thresh_ms = 1000

        self.playback_rate = 1
        self.enabled_at = None
        self.ready = False
        self.is_buffering = False

        self.last_command = None
        self.queued_command = None

        self.scheduled_command = None
        self.sync_timeout = None

        self.time_offset = timedelta(0)
        self.round_trip_duration = timedelta(0)
        self.notify_sync_ready = False

        self.read_callback = None
        self.timesync = None
        self.client = None
        self.current_group = None
        self.playqueue_last_updated = None

    # On playback time update (call from timeline push)
    def sync_playback_time(self):
        if (
            not self.last_command
            or self.last_command["Command"] != "Unpause"
            or self.is_buffering
            or not self.playerManager.is_not_paused()
            or self.playerManager.menu.is_menu_shown
        ):
            log.debug("Not syncing due to no playback.")
            return

        current_time = datetime.utcnow()

        # Avoid overloading player
        elapsed = current_time - self.last_sync_time
        if elapsed.total_seconds() * 1000 < settings.sync_method_thresh / 2:
            log.debug("Not syncing due to threshold.")
            return

        self.last_sync_time = current_time
        play_at_time = self.last_command["When"]

        current_position_ticks = int(self.playerManager.get_time() * seconds_in_ticks)
        server_position_ticks = (
            self.last_command["PositionTicks"]
            + ((current_time - play_at_time) + self.time_offset).total_seconds()
            * seconds_in_ticks
        )

        diff_ms = (server_position_ticks - current_position_ticks) / 10000
        self.playback_diff_ms = diff_ms

        if self.sync_enabled:
            abs_diff_ms = abs(diff_ms)

            if (
                self.enable_speed_sync
                and settings.sync_max_delay_speed
                < abs_diff_ms
                < settings.sync_method_thresh
            ):
                if self.attempts > settings.sync_speed_attempts:
                    self.enable_speed_sync = False
                    return

                # Speed To Sync Method
                speed = 1 + diff_ms / settings.sync_speed_time

                self.playerManager.set_speed(speed)
                self.sync_enabled = False
                self.attempts += 1
                log.info("SyncPlay Speed to Sync rate: {0}".format(speed))
                self.player_message(_("SpeedToSync (x{0})").format(speed))

                def callback():
                    self.playerManager.set_speed(1)
                    self.sync_enabled = True

                set_timeout(settings.sync_speed_time, callback)
            elif abs_diff_ms > settings.sync_max_delay_skip:
                if self.attempts > settings.sync_attempts:
                    self.sync_enabled = False
                    log.info("SyncPlay Sync Disabled due to too many attempts.")
                    self.player_message(_("Sync Disabled (Too Many Attempts)"))
                    return

                # Skip To Sync Method
                self.local_seek(server_position_ticks / seconds_in_ticks)
                self.sync_enabled = False
                self.attempts += 1
                log.info("SyncPlay Skip to Sync Activated")
                self.player_message(_("SkipToSync (x{0})").format(self.attempts))

                def callback():
                    self.sync_enabled = True

                set_timeout(settings.sync_method_thresh / 2, callback)
            else:
                if self.attempts > 0:
                    log.info(
                        "Playback synced after {0} attempts.".format(self.attempts)
                    )
                self.attempts = 0

    # On timesync update
    def on_timesync_update(self, time_offset: timedelta, ping: timedelta):
        self.time_offset = time_offset
        self.round_trip_duration = ping * 2

        if self.notify_sync_ready:
            self.ready = True
            self.notify_sync_ready = False
            if self.read_callback:
                self.read_callback()
                self.read_callback = None

        # Server responds with 400 bad request...
        if self.sync_enabled:
            try:
                self.client.jellyfin.ping_sync_play(ping.total_seconds() * 1000)
            except Exception:
                log.error("Syncplay ping reporting failed.")

    def enable_sync_play(self, from_server: bool):
        self.playback_rate = self.playerManager.get_speed()
        self.enabled_at = datetime.utcnow()
        self.enable_speed_sync = True

        def ready_callback():
            self.process_command(self.queued_command)
            self.queued_command = None
            self.enabled_at = self.timesync.local_date_to_server(self.enabled_at)

        self.read_callback = ready_callback

        self.ready = False
        self.notify_sync_ready = True

        if not from_server:
            self.client = self.playerManager.get_current_client()

        timesync = self.client.timesync
        if self.timesync is not None and timesync is not self.timesync:
            self.timesync.remove_subscriber(self.on_timesync_update)
            self.timesync.stop_ping()

        self.timesync = timesync
        self.timesync.subscribe_time_offset(self.on_timesync_update)
        self.timesync.force_update()

        log.info("Syncplay enabled.")
        if from_server:
            self.player_message(_("SyncPlay enabled."))

    def disable_sync_play(self, from_server: bool):
        self.playerManager.set_speed(self.playback_rate)

        self.enabled_at = None
        self.ready = False
        self.last_command = None
        self.queued_command = None
        self.sync_enabled = False
        self.playqueue_last_updated = None

        if self.timesync is not None:
            self.timesync.remove_subscriber(self.on_timesync_update)
            self.timesync.stop_ping()
            self.timesync = None
        self.current_group = None

        log.info("Syncplay disabled.")
        if from_server:
            self.player_message(_("SyncPlay disabled."))
        else:
            try:
                self.client.jellyfin.leave_sync_play()
            except:
                log.warning("Failed to leave syncplay.", exc_info=True)

    def _buffer_req(self, is_buffering):
        if self.timesync is None:
            # This can get called before it is ready...
            return
        media = self.playerManager.get_video().parent
        when = self.timesync.local_date_to_server(datetime.utcnow())
        ticks = int(self.playerManager.get_time() * seconds_in_ticks)
        playing = self.playerManager.is_not_paused()
        playlist_id = media.queue[media.seq]["PlaylistItemId"]

        if is_buffering:
            self.client.jellyfin.buffering_sync_play(when, ticks, playing, playlist_id)
        else:
            self.client.jellyfin.ready_sync_play(when, ticks, playing, playlist_id)

    # On Buffer
    def on_buffer(self):
        if not self.last_playback_waiting:
            playback_waiting = datetime.utcnow()
            self.last_playback_waiting = playback_waiting

            def handle_buffer():
                if playback_waiting == self.last_playback_waiting:
                    self._buffer_req(True)
                    self.is_buffering = True

            set_timeout(self.min_buffer_thresh_ms, handle_buffer)

    # On Buffer Done
    def on_buffer_done(self):
        if self.is_buffering:
            self._buffer_req(False)

        self.last_playback_waiting = None
        self.is_buffering = False

    def play_done(self):
        self.local_pause()
        self._buffer_req(False)

    def is_enabled(self):
        return self.enabled_at is not None

    def process_group_update(self, command: dict):
        command_type = command["Type"]
        log.debug("Syncplay group update: {0}".format(command))
        if command_type in info_commands:
            self.player_message(info_commands[command_type])
        elif command_type == "PrepareSession":
            self.prepare_session(command["GroupId"], command["Data"])
        elif command_type == "GroupJoined":
            self.current_group = command["GroupId"]
            self.enable_sync_play(True)
        elif command_type == "GroupLeft" or command_type == "NotInGroup":
            self.disable_sync_play(True)
        elif command_type == "UserJoined":
            self.player_message(_("{0} has joined.").format(command["Data"]))
        elif command_type == "UserLeft":
            self.player_message(_("{0} has left.").format(command["Data"]))
        elif command_type == "GroupWait":
            self.player_message(_("{0} is buffering.").format(command["Data"]))
        elif command_type == "PlayQueue":
            self.upd_queue(command["Data"])
        elif command_type == "StateUpdate":
            log.info(
                "{0} state caused by {1}".format(
                    command["Data"]["State"], command["Data"]["Reason"]
                )
            )
        else:
            log.error(
                "Unknown SyncPlay command {0} payload {1}.".format(
                    command_type, command
                )
            )

    def process_command(self, command: Optional[dict]):
        if command is None:
            return

        if not self.is_enabled():
            log.debug(
                "Ignoring command {0} due to SyncPlay being disabled.".format(command)
            )
            return

        if not self.ready:
            log.debug(
                "Queued command {0} due to SyncPlay not being ready.".format(command)
            )
            self.queued_command = command
            return

        command["When"] = _parse_precise_time(command["When"])
        command["EmitttedAt"] = _parse_precise_time(command["EmittedAt"])

        if command["EmitttedAt"] < self.enabled_at:
            log.debug("Ignoring old command {0}.".format(command))
            return

        if (
            self.last_command
            and self.last_command["When"] == command["When"]
            and self.last_command["PositionTicks"] == command["PositionTicks"]
            and self.last_command["Command"] == command["Command"]
        ):
            log.debug("Ignoring duplicate command {0}.".format(command))
            return

        self.last_command = command
        command_cmd, when, position = (
            command["Command"],
            command["When"],
            command["PositionTicks"],
        )
        log.info(
            "Syncplay will {0} at {1} position {2}".format(command_cmd, when, position)
        )

        if command_cmd == "Unpause":
            self.schedule_play(when, position)
        elif command_cmd == "Pause":
            self.schedule_pause(when, position)
        elif command_cmd == "Seek":
            self.schedule_seek(when, position)
        else:
            log.error("Command {0} is unknown.".format(command_cmd))

    def prepare_session(self, group_id: str, session_data: dict):
        # I think this might be a dead code path
        play_command = session_data.get("PlayCommand")
        if not self.playerManager.has_video():
            play_command = "PlayNow"

        seq = session_data.get("StartIndex")
        if seq is None:
            seq = 0
        media = Media(
            self.client,
            session_data.get("ItemIds"),
            seq=seq,
            user_id=session_data.get("ControllingUserId"),
            aid=session_data.get("AudioStreamIndex"),
            sid=session_data.get("SubtitleStreamIndex"),
            srcid=session_data.get("MediaSourceId"),
        )

        if (
            self.playerManager.has_video()
            and self.playerManager.get_video().item_id == session_data["ItemIds"][0]
            and play_command == "PlayNow"
        ):
            # We assume the video is already available.
            self.playerManager.get_video().parent = media
            log.info("Syncplay Session Prepare: {0} {1}".format(group_id, session_data))
            self.local_seek(
                (session_data.get("PositionTicks", 0) or 0) / seconds_in_ticks
            )
            self.current_group = group_id
        elif play_command == "PlayNow":
            log.info(
                "Syncplay Session Recreate: {0} {1}".format(group_id, session_data)
            )
            offset = session_data.get("StartPositionTicks")
            if offset is not None:
                offset /= 10000000

            video = media.video
            if video:
                if settings.pre_media_cmd:
                    os.system(settings.pre_media_cmd)
                self.playerManager.play(
                    video, offset, no_initial_timeline=True, is_initial_play=True
                )
                self.playerManager.send_timeline()
                # Really not sure why I have to call this.
                self.join_group(group_id)
                self.playerManager.timeline_handle()
                if settings.play_cmd:
                    os.system(settings.play_cmd)
        elif play_command == "PlayLast":
            self.playerManager.get_video().parent.insert_items(
                session_data.get("ItemIds"), append=True
            )
            self.playerManager.upd_player_hide()
        elif play_command == "PlayNext":
            self.playerManager.get_video().parent.insert_items(
                session_data.get("ItemIds"), append=False
            )
            self.playerManager.upd_player_hide()

    def player_message(self, message: str):
        # Messages overwrite menu, so they are ignored.
        if not self.menu.is_menu_shown:
            if settings.sync_osd_message:
                self.playerManager.show_text(message, 2000)
            else:
                log.info("SyncPlay Message: {0}".format(message))
        else:
            log.info("Ignored SyncPlay Message (menu): {0}".format(message))

    def schedule_play(self, when: datetime, position: int):
        self.clear_scheduled_command()
        current_time = datetime.utcnow()
        local_play_time = self.timesync.server_date_to_local(when)

        if local_play_time > current_time:
            log.debug("SyncPlay Scheduled Play: Playing Later")
            play_timeout = (local_play_time - current_time).total_seconds() * 1000
            self.local_seek(position / seconds_in_ticks)

            def scheduled():
                self.local_play()

                def sync_timeout():
                    self.sync_enabled = True

                self.sync_timeout = set_timeout(
                    settings.sync_method_thresh / 2, sync_timeout
                )

            self.scheduled_command = set_timeout(play_timeout, scheduled)
        else:
            log.debug("SyncPlay Scheduled Play: Playing Now")
            # Group playback already started
            server_position_secs = (
                position / seconds_in_ticks
                + (current_time - local_play_time).total_seconds()
            )
            self.local_play()
            self.local_seek(server_position_secs)

            def sync_timeout():
                self.sync_enabled = True

            self.sync_timeout = set_timeout(
                settings.sync_method_thresh / 2, sync_timeout
            )

    def schedule_pause(self, when: datetime, position: int, seek_only: bool = False):
        self.clear_scheduled_command()
        current_time = datetime.utcnow()
        local_pause_time = self.timesync.server_date_to_local(when)

        def callback():
            if not seek_only:
                self.local_pause()
            self.local_seek(position / seconds_in_ticks)

        if local_pause_time > current_time:
            log.debug("SyncPlay Scheduled Pause/Seek: Pausing Later")
            pause_timeout = (local_pause_time - current_time).total_seconds() * 1000
            self.scheduled_command = set_timeout(pause_timeout, callback)
        else:
            log.debug("SyncPlay Scheduled Pause/Seek: Pausing Now")
            callback()

    def _play_video(self, video, offset):
        if video:
            if settings.pre_media_cmd:
                os.system(settings.pre_media_cmd)
            self.playerManager.play(video, offset, no_initial_timeline=True)
            if settings.play_cmd:
                os.system(settings.play_cmd)
        else:
            log.error("No video from queue update.")

    def upd_queue(self, data):
        # It can't hurt to update the queue lol.
        # last_upd = _parse_precise_time(data["LastUpdate"])
        # if (
        #    self.playqueue_last_updated is not None
        #    and self.playqueue_last_updated >= last_upd
        # ):
        #    log.warning("Tried to apply old queue update.")
        #    return
        #
        # self.playqueue_last_updated = last_upd

        sp_items = [
            {"Id": x["ItemId"], "PlaylistItemId": x["PlaylistItemId"]}
            for x in data["Playlist"]
        ]
        if self.playerManager.get_video() is None:
            media = Media(
                self.client, sp_items, data["PlayingItemIndex"], queue_override=False
            )
            log.info("The queue update changed the video. (New)")
            offset = data.get("StartPositionTicks")
            if offset is not None:
                offset /= 10000000

            self._play_video(media.video, offset)
        else:
            media = self.playerManager.get_video().parent
            new_media = media.replace_queue(sp_items, data["PlayingItemIndex"])
            if new_media:
                log.info("The queue update changed the video.")
                offset = data.get("StartPositionTicks")
                if offset is not None:
                    offset /= 10000000

                self._play_video(new_media.video, offset)
            else:
                self._buffer_req(False)

    def schedule_seek(self, when: datetime, position: int):
        # This replicates what the web client does.
        self.schedule_pause(when, position)

    def clear_scheduled_command(self):
        if self.scheduled_command is not None:
            self.scheduled_command()

        if self.sync_timeout is not None:
            self.sync_timeout()

        self.sync_enabled = False
        self.playerManager.set_speed(1)

    def play_request(self):
        self.client.jellyfin.unpause_sync_play()

    def pause_request(self):
        self.client.jellyfin.pause_sync_play()
        self.playerManager.set_paused(True, True)

    def seek_request(self, offset: float):
        self.client.jellyfin.seek_sync_play(int(offset * seconds_in_ticks))

    def local_play(self):
        self.playerManager.set_paused(False, True)

    def local_pause(self):
        self.playerManager.set_paused(True, True)

    def local_seek(self, offset: float):
        self.playerManager.seek(offset, absolute=True, force=True)

    def join_group(self, group_id: str):
        self.client.jellyfin.join_sync_play(group_id)

    def discord_join_group(self, join_secret):
        self.enable_sync_play(False)
        self.client.jellyfin.join_sync_play(join_secret["secret"])

    def menu_join_group(self):
        group = self.menu.menu_list[self.menu.menu_selection][2]
        self.menu.hide_menu()
        self.client.jellyfin.join_sync_play(group["GroupId"])

    def menu_disable(self):
        self.menu.hide_menu()
        self.client.jellyfin.leave_sync_play()

    def menu_create_group(self):
        self.menu.hide_menu()
        self.client.jellyfin.new_sync_play_v2(
            _("{0}'s Group").format(clientManager.get_username_from_client(self.client))
        )

    def menu_action(self):
        self.client = self.playerManager.get_current_client()

        selected = 0
        offset = 1
        group_option_list = [
            (_("None (Disabled)"), self.menu_disable, None),
        ]
        if not self.is_enabled():
            offset = 2
            group_option_list.append((_("New Group"), self.menu_create_group, None))
        groups = self.client.jellyfin.get_sync_play()
        for i, group in enumerate(groups):
            group_option_list.append((group["GroupName"], self.menu_join_group, group))
            if group["GroupId"] == self.current_group:
                selected = i + offset
        self.menu.put_menu(_("SyncPlay"), group_option_list, selected)

    def request_next(self, playlist_item_id):
        self.client.jellyfin.next_sync_play(playlist_item_id)

    def request_prev(self, playlist_item_id):
        self.client.jellyfin.prev_sync_play(playlist_item_id)

    def request_skip(self, playlist_item_id):
        self.client.jellyfin.set_item_sync_play(playlist_item_id)



import logging
import os
import sys
import time
import platform

from threading import RLock, Lock, Event
from queue import Queue
from collections import OrderedDict
from typing import TYPE_CHECKING, Optional

from . import conffile
from .utils import synchronous, Timer, none_fallback, get_resource
from .conf import settings
from .menu import OSDMenu
from .constants import APP_NAME
from .syncplay import SyncPlayManager
from .update_check import UpdateChecker
from .i18n import _

if TYPE_CHECKING:
    from .media import Video as Video_type

log = logging.getLogger("player")
mpv_log = logging.getLogger("mpv")


discord_presence = False
if settings.discord_presence:
    try:
        from .rich_presence import register_join_event, send_presence, clear_presence

        discord_presence = True
    except Exception:
        log.error("Could not enable Discord Rich Presence.", exc_info=True)

python_mpv_available = True
is_using_ext_mpv = False
if not settings.mpv_ext:
    try:
        # noinspection PyPackageRequirements
        import mpv

        log.info("Using libmpv playback backend.")
    except OSError:
        log.warning("Could not find libmpv.")
        python_mpv_available = False

if settings.mpv_ext or not python_mpv_available:
    import python_mpv_jsonipc as mpv

    log.info("Using external mpv playback backend.")
    is_using_ext_mpv = True

SUBTITLE_POS = {
    "top": 0,
    "bottom": 100,
    "middle": 80,
}

mpv_log_levels = {
    "fatal": mpv_log.error,
    "error": mpv_log.error,
    "warn": mpv_log.warning,
    "info": mpv_log.info,
}


def mpv_log_handler(level: str, prefix: str, text: str):
    if level in mpv_log_levels:
        mpv_log_levels[level]("{0}: {1}".format(prefix, text))
    else:
        mpv_log.debug("{0}: {1}".format(prefix, text))


def wait_property(
    instance, name: str, cond=lambda x: True, timeout: Optional[int] = None
):
    success = True
    event = Event()

    def handler(_name, value):
        if cond(value):
            event.set()

    if is_using_ext_mpv:
        observer_id = instance.bind_property_observer(name, handler)
        if timeout:
            success = event.wait(timeout=timeout)
        else:
            event.wait()
        instance.unbind_property_observer(observer_id)
    else:
        instance.observe_property(name, handler)
        if timeout:
            success = event.wait(timeout=timeout)
        else:
            event.wait()
        instance.unobserve_property(name, handler)
    return success


win_utils = None
if sys.platform.startswith("win32") or sys.platform.startswith("cygwin"):
    try:
        from . import win_utils
    except ModuleNotFoundError:
        log.warning("win_utils is not available.")

# Q: What is with the put_task call?
# A: Some calls to python-mpv require event processing.
#    put_task is used to deal with the events originating from
#    the event thread, which would cause deadlock if they run there.


# noinspection PyUnresolvedReferences
class PlayerManager(object):
    """
    The underlying player is thread safe, however, locks are used in this
    class to prevent concurrent control events hitting the player, which
    violates assumptions.
    """

    def __init__(self):
        self._video = None
        mpv_options = OrderedDict()
        mpv_location = settings.mpv_ext_path
        # Use bundled path for MPV if not specified by user, on Mac OS, and frozen
        if (
            mpv_location is None
            and platform.system() == "Darwin"
            and getattr(sys, "frozen", False)
        ):
            mpv_location = get_resource("mpv")
        self.timeline_trigger = None
        self.action_trigger = None
        self.external_subtitles = {}
        self.external_subtitles_rev = {}
        self.should_send_timeline = False
        self.start_time = None
        self.url = None
        self.evt_queue = Queue()
        self._lock = RLock()
        self._tl_lock = RLock()
        self._finished_lock = Lock()
        self.last_update = Timer()
        self._jf_settings = None
        self.get_webview = lambda: None
        self.pause_ignore = None  # Used to ignore pause events that come from us.
        self.do_not_handle_pause = False
        self.last_seek = None
        self.warned_about_transcode = False
        self.fullscreen_disable = False
        self.update_check = UpdateChecker(self)
        self.is_in_intro = False
        self.trickplay = None

        if is_using_ext_mpv:
            mpv_options.update(
                {
                    "start_mpv": settings.mpv_ext_start,
                    "ipc_socket": settings.mpv_ext_ipc,
                    "mpv_location": mpv_location,
                    "player-operation-mode": "cplayer",
                }
            )

        scripts = []
        if settings.menu_mouse:
            scripts.append(get_resource("mouse.lua"))
        if settings.thumbnail_enable:
            try:
                from .trickplay import TrickPlay

                self.trickplay = TrickPlay(self)
                self.trickplay.start()

                scripts.append(get_resource("thumbfast.lua"))
                if settings.thumbnail_osc_builtin:
                    scripts.append(get_resource("trickplay-osc.lua"))

            except Exception:
                log.error("Could not enable trickplay.", exc_info=True)

            mpv_options["osc"] = False

        # ensure standard mpv configuration directories and files exist
        conffile.get_dir(APP_NAME, "scripts")
        conffile.get_dir(APP_NAME, "fonts")
        conffile.get(APP_NAME, "input.conf", True)
        conffile.get(APP_NAME, "mpv.conf", True)

        if scripts:
            if settings.mpv_ext:
                mpv_options["script"] = scripts
            else:
                mpv_options["scripts"] = (
                    ";" if sys.platform.startswith("win32") else ":"
                ).join(scripts)

        if not (settings.mpv_ext and settings.mpv_ext_no_ovr):
            mpv_options["config"] = True
            mpv_options["config_dir"] = conffile.confdir(APP_NAME)

        if settings.tls_client_cert and settings.tls_client_key:
            mpv_options['tls_cert_file'] = settings.tls_client_cert
            mpv_options['tls_key_file'] = settings.tls_client_key

            if settings.tls_server_ca:
                mpv_options['tls_ca_file'] = settings.tls_server_ca

        self._player = mpv.MPV(
            input_default_bindings=True,
            input_vo_keyboard=True,
            input_media_keys=settings.media_keys,
            log_handler=mpv_log_handler,
            loglevel=settings.mpv_log_level,
            **mpv_options,
        )

        self.menu = OSDMenu(self, self._player)
        self.syncplay = SyncPlayManager(self)

        if discord_presence:
            try:
                register_join_event(self.syncplay.discord_join_group)
            except Exception:
                log.error("Could not register Discord join callback.", exc_info=True)

        if hasattr(self._player, "osc"):
            self.enable_osc(settings.enable_osc)
        else:
            log.warning("This mpv version doesn't support on-screen controller.")

        if settings.screenshot_dir is not None:
            if hasattr(self._player, "screenshot_directory"):
                self._player.screenshot_directory = settings.screenshot_dir
            else:
                log.warning(
                    "This mpv version doesn't support setting the screenshot directory."
                )

        if hasattr(self._player, "resume_playback"):
            # This can lead to unwanted skipping of videos
            self._player.resume_playback = False

        # Wrapper for on_key_press that ignores None.
        def keypress(key):
            def wrapper(func):
                if key is not None:
                    self._player.on_key_press(key)(func)
                return func

            return wrapper

        @self._player.on_key_press("CLOSE_WIN")
        @self._player.on_key_press("STOP")
        @keypress(settings.kb_stop)
        def handle_stop():
            self.stop()

        @keypress(settings.kb_prev)
        def handle_prev():
            self.put_task(self.play_prev)

        @keypress(settings.kb_next)
        def handle_next():
            self.put_task(self.play_next)

        @self._player.on_key_press("PREV")
        @self._player.on_key_press("XF86_PREV")
        def handle_media_prev():
            if settings.media_key_seek:
                seektime, _x = self.get_seek_times()
                self.seek(seektime)
            else:
                self.put_task(self.play_prev)

        @self._player.on_key_press("NEXT")
        @self._player.on_key_press("XF86_NEXT")
        def handle_media_next():
            if settings.media_key_seek:
                if self.is_in_intro:
                    self.skip_intro()
                else:
                    _x, seektime = self.get_seek_times()
                    self.seek(seektime)
            else:
                self.put_task(self.play_next)

        @keypress(settings.kb_watched)
        def handle_watched():
            self.put_task(self.watched_skip)

        @keypress(settings.kb_unwatched)
        def handle_unwatched():
            self.put_task(self.unwatched_quit)

        @keypress(settings.kb_menu)
        def menu_open():
            if not self.menu.is_menu_shown:
                self.menu.show_menu()
            else:
                self.menu.hide_menu()

        @keypress(settings.kb_menu_esc)
        def menu_back():
            if self.menu.is_menu_shown:
                self.menu.menu_action("back")
            else:
                self._player.command("set", "fullscreen", "no")
                self.fullscreen_disable = True

        @keypress(settings.kb_menu_ok)
        def menu_ok():
            self.menu.menu_action("ok")

        @keypress(settings.kb_menu_left)
        def menu_left():
            if self.menu.is_menu_shown:
                self.menu.menu_action("left")
            else:
                self.kb_seek("left")

        @keypress(settings.kb_menu_right)
        def menu_right():
            if self.menu.is_menu_shown:
                self.menu.menu_action("right")
            else:
                if self.is_in_intro:
                    self.skip_intro()
                else:
                    self.kb_seek("right")

        @keypress(settings.kb_menu_up)
        def menu_up():
            if self.menu.is_menu_shown:
                self.menu.menu_action("up")
            else:
                if self.is_in_intro:
                    self.skip_intro()
                else:
                    self.kb_seek("up")

        @keypress(settings.kb_menu_down)
        def menu_down():
            if self.menu.is_menu_shown:
                self.menu.menu_action("down")
            else:
                self.kb_seek("down")

        @keypress(settings.kb_pause)
        def handle_pause():
            if self.menu.is_menu_shown:
                self.menu.menu_action("ok")
            else:
                self.toggle_pause()

        @keypress(settings.kb_fullscreen)
        def handle_fullscreen():
            self.toggle_fullscreen()

        # This gives you an interactive python debugger prompt.
        @keypress(settings.kb_debug)
        def handle_debug():
            import pdb

            pdb.set_trace()

        # Kill shader packs (useful for breakage)
        @keypress(settings.kb_kill_shader)
        def kill_shaders():
            if settings.shader_pack_remember:
                settings.shader_pack_profile = None
                settings.save()
            if self.menu.profile_manager is not None:
                self.menu.profile_manager.unload_profile()

        # Fires between episodes.
        @self._player.property_observer("eof-reached")
        def handle_end(_name, reached_end: bool):
            self.pause_ignore = True
            if self._video and reached_end:
                has_lock = self._finished_lock.acquire(False)
                self.put_task(self.finished_callback, has_lock)

        # Fires at the end.
        @self._player.property_observer("playback-abort")
        def handle_end_idle(_name, value: bool):
            self.pause_ignore = True
            if self._video and value and not self._video.parent.has_next:
                has_lock = self._finished_lock.acquire(False)
                self.put_task(self.finished_callback, has_lock)

        @self._player.property_observer("seeking")
        def handle_seeking(_name, value: bool):
            if self.do_not_handle_pause:
                return

            if self.syncplay.is_enabled():
                play_time = self._player.playback_time
                if (
                    play_time is not None
                    and self.last_seek is not None
                    and abs(self.last_seek - play_time) > 10
                ):
                    self.syncplay.seek_request(play_time)
                else:
                    log.info("SyncPlay Buffering: {0}".format(value))
                    if value:
                        self.syncplay.on_buffer()
                    else:
                        self.syncplay.on_buffer_done()

        @self._player.property_observer("pause")
        def pause_handler(_name, value: bool):
            if self.do_not_handle_pause:
                return

            if not self._player.playback_abort:
                self.timeline_handle()

            if value != self.pause_ignore:
                if self.syncplay.is_enabled():
                    if value:
                        self.syncplay.pause_request()
                    else:
                        # Don't allow unpausing locally through MPV.
                        self.syncplay.play_request()
                        self.set_paused(True, True)

        @self._player.event_callback("client-message")
        def handle_client_message(event):
            try:
                # Python-MPV 1.0 uses a class/struct combination now
                if hasattr(event, "as_dict"):
                    event = event.as_dict()
                    if "event" in event:
                        event["event"] = event["event"].decode("utf-8")
                    if "args" in event:
                        event["args"] = [d.decode("utf-8") for d in event["args"]]

                if "event_id" in event:
                    args = event["event"]["args"]
                else:
                    args = event["args"]
                if len(args) == 0:
                    return
                if args[0] == "shim-menu-select":
                    # Apparently this can happen...
                    if args[1] == "inf":
                        return
                    self.menu.mouse_select(int(args[1]))
                elif args[0] == "shim-menu-click":
                    self.menu.menu_action("ok")
            except Exception:
                log.warning("Error when processing client-message.", exc_info=True)

    # Put a task to the event queue.
    # This ensures the task executes outside
    # of an event handler, which causes a crash.
    def put_task(self, func, *args):
        self.evt_queue.put([func, args])
        if self.action_trigger:
            self.action_trigger.set()

    # Trigger the timeline to update all
    # clients immediately.
    def timeline_handle(self):
        if self.timeline_trigger:
            self.timeline_trigger.set()

    def skip_intro(self):
        _, intro = self._video.get_current_intro(self._player.playback_time)

        self._player.playback_time = intro.end
        intro.has_triggered = True
        self.timeline_handle()
        self.is_in_intro = False

    @synchronous("_lock")
    def update(self):
        if (
            (
                settings.skip_intro_always
                or settings.skip_intro_enable
                or settings.skip_credits_always
                or settings.skip_credits_enable
            )
            and not self.syncplay.is_enabled()
            and self._video is not None
            and self._player.playback_time is not None
        ):
            ready_to_skip, intro = self._video.get_current_intro(
                self._player.playback_time
            )

            if intro is not None:
                should_prompt = (
                    intro.type != "Outro" and settings.skip_intro_enable
                ) or (intro.type == "Outro" and settings.skip_credits_enable)
                should_skip = (not intro.has_triggered) and (
                    (intro.type != "Outro" and settings.skip_intro_always)
                    or (intro.type == "Outro" and settings.skip_credits_always)
                )

                if should_skip and ready_to_skip:
                    intro.has_triggered = True
                    self.skip_intro()
                    self._player.show_text(
                        _("Skipped Credits")
                        if intro.type == "Outro"
                        else _("Skipped Intro"),
                        3000,
                        1,
                    )

                if not self.is_in_intro and should_prompt:
                    self._player.show_text(
                        _("Seek to Skip Credits")
                        if intro.type == "Outro"
                        else _("Seek to Skip Intro"),
                        3000,
                        1,
                    )
                self.is_in_intro = True
            else:
                self.is_in_intro = False

        while not self.evt_queue.empty():
            func, args = self.evt_queue.get()
            func(*args)
        if self._video and not self._player.playback_abort:
            if not self.is_paused():
                self.last_update.restart()

    def play(
        self,
        video: "Video_type",
        offset: int = 0,
        no_initial_timeline: bool = False,
        is_initial_play: bool = False,
    ):
        self.should_send_timeline = False
        self.start_time = time.time()
        url = video.get_playback_url()
        if not url:
            log.error("PlayerManager::play no URL found")
            return

        self._play_media(video, url, offset, no_initial_timeline, is_initial_play)

    @synchronous("_lock")
    def _play_media(
        self,
        video: "Video_type",
        url: str,
        offset: int = 0,
        no_initial_timeline: bool = False,
        is_initial_play: bool = False,
    ):
        self.pause_ignore = True
        self.do_not_handle_pause = True
        self.url = url
        self.menu.hide_menu()

        if self.trickplay:
            self.trickplay.clear()

        if settings.log_decisions:
            log.info("Playing: {0}".format(url))
        if self.get_webview() is not None and settings.display_mirroring:
            # noinspection PyUnresolvedReferences
            self.get_webview().hide()

        self._player.play(self.url)
        if not wait_property(
            self._player, "duration", lambda x: x is not None, settings.playback_timeout
        ):
            # Timeout playback attempt after 10 seconds
            log.error("Timeout when waiting for media duration. Stopping playback!")
            self.stop()
            return
        log.info("Finished waiting for media duration.")
        if settings.fullscreen and not self.fullscreen_disable:
            self._player.fs = True
        self._player.force_media_title = video.get_proper_title()
        self._video = video
        self.is_in_intro = False
        self.external_subtitles = {}
        self.external_subtitles_rev = {}

        self.upd_player_hide()
        self.configure_streams()
        self.update_subtitle_visuals()

        if win_utils and settings.raise_mpv and is_initial_play:
            win_utils.raise_mpv()

        if offset is not None and offset > 0:
            self.last_seek = offset
            self._player.playback_time = offset

        if not no_initial_timeline:
            self.send_timeline_initial()
        else:
            self.send_timeline()

        if self.syncplay.is_enabled():
            self.set_speed(1)
            self.syncplay.play_done()
        else:
            self.set_paused(False, False)

        if self.trickplay:
            self.trickplay.fetch_thumbnails()

        self.should_send_timeline = True
        self.do_not_handle_pause = False
        if self._finished_lock.locked():
            self._finished_lock.release()

        self.update_check.check()

        if (
            not self._video.parent.is_local
            and self._video.is_transcode
            and not self.warned_about_transcode
            and settings.transcode_warning
        ):
            self.warned_about_transcode = True
            self._player.show_text(
                _(
                    "Your remote video is transcoding!\nPress c to adjust bandwidth settings if this is not needed."
                ),
                5000,
                1,
            )

    @staticmethod
    def exec_stop_cmd():
        if settings.stop_cmd:
            os.system(settings.stop_cmd)

    @synchronous("_lock")
    def stop(self):
        if self.syncplay.is_enabled():
            self.syncplay.disable_sync_play(False)

        if not self._video or self._player.playback_abort:
            self.exec_stop_cmd()
            return

        log.info("PlayerManager::stop stopping playback of %s" % self._video)

        self.should_send_timeline = False
        options = self.get_timeline_options()
        self.set_paused(False)
        local_video = self._video
        self._video = None
        self._player.command("stop")
        local_video.terminate_transcode()
        self.send_timeline_stopped(options=options, client=local_video.client)
        self.exec_stop_cmd()

        if self.trickplay:
            self.trickplay.clear()

    def get_volume(self, percent: bool = False):
        if self._player:
            if not percent:
                return self._player.volume / 100
            return self._player.volume

    @synchronous("_lock")
    def toggle_pause(self):
        if not self._player.playback_abort:
            self.set_paused(not self._player.pause)

    @synchronous("_lock")
    def pause_if_playing(self):
        if not self._player.playback_abort:
            if not self._player.pause:
                self.set_paused(True)
        self.timeline_handle()

    @synchronous("_lock")
    def play_if_paused(self):
        if not self._player.playback_abort:
            if self._player.pause:
                self.set_paused(False)
        self.timeline_handle()

    @synchronous("_lock")
    def seek(
        self,
        offset: float,
        absolute: bool = False,
        force: bool = False,
        exact: Optional[bool] = None,
    ):
        """
        Seek to ``offset`` seconds
        """
        if exact is None:
            exact = absolute
        if self.syncplay.is_enabled() and not force:
            if not absolute:
                offset += self._player.playback_time
            self.syncplay.seek_request(offset)
        else:
            if not self._player.playback_abort:
                if absolute:
                    if self.syncplay.is_enabled():
                        self.last_seek = offset
                    if self.is_in_intro and offset > self._player.playback_time:
                        self.skip_intro()
                    p2 = "absolute"
                    if exact:
                        p2 += "+exact"
                    self._player.command("seek", offset, p2)
                else:
                    if self.syncplay.is_enabled():
                        self.last_seek = self._player.playback_time + offset
                    if (
                        self.is_in_intro
                        and self._player.playback_time + offset
                        > self._player.playback_time
                    ):
                        self.skip_intro()
                    if exact:
                        self._player.command("seek", offset, "exact")
                    else:
                        self._player.command("seek", offset)
        self.timeline_handle()

    @synchronous("_lock")
    def set_volume(self, pct: float):
        if not self._player.playback_abort:
            self._player.volume = pct
        self.timeline_handle()

    @synchronous("_lock")
    def get_state(self):
        if self._player.playback_abort:
            return "stopped"

        if self._player.pause:
            return "paused"

        return "playing"

    @synchronous("_lock")
    def is_paused(self):
        if not self._player.playback_abort:
            return self._player.pause
        return False

    @synchronous("_lock")
    def finished_callback(self, has_lock: bool):
        if not self._video:
            self.pause_ignore = False
            return

        if settings.force_set_played:
            self._video.set_played()
        if self._video.parent.has_next and settings.auto_play:
            if has_lock:
                log.info("PlayerManager::finished_callback starting next episode")
                new_video = self._video.parent.get_next().video
                self.send_timeline_stopped(True)
                if self.syncplay.is_enabled():
                    self.syncplay.request_next(self._video.get_playlist_id())
                else:
                    self.play(new_video)
            else:
                log.info("PlayerManager::finished_callback No lock, skipping...")
        else:
            if settings.media_ended_cmd:
                os.system(settings.media_ended_cmd)

            if self.syncplay.is_enabled():
                self.syncplay.disable_sync_play(False)

            log.info("PlayerManager::finished_callback reached end")
            self.send_timeline_stopped(True)
        self.pause_ignore = False

    @synchronous("_lock")
    def watched_skip(self):
        if not self._video:
            return

        self._video.set_played()
        self.play_next()

    @synchronous("_lock")
    def unwatched_quit(self):
        if not self._video:
            return

        video = self._video
        self.stop()
        video.set_played(False)

    @synchronous("_lock")
    def play_next(self):
        if self._video.parent.has_next:
            new_video = self._video.parent.get_next().video
            self.send_timeline_stopped(True)
            if self.syncplay.is_enabled():
                self.syncplay.request_next(self._video.get_playlist_id())
            else:
                self.play(new_video)
            return True
        return False

    @synchronous("_lock")
    def skip_to(self, key: str):
        media = self._video.parent.get_from_key(key)
        if media:
            self.send_timeline_stopped(True)
            if self.syncplay.is_enabled():
                self.syncplay.request_skip(media.video.get_playlist_id())
            else:
                self.play(media.get_video(0))
            return True
        return False

    @synchronous("_lock")
    def play_prev(self):
        if self._video.parent.has_prev:
            new_video = self._video.parent.get_prev().video
            self.send_timeline_stopped(True)
            if self.syncplay.is_enabled():
                self.syncplay.request_prev(self._video.get_playlist_id())
            else:
                self.play(new_video)
            return True
        return False

    @synchronous("_lock")
    def restart_playback(self):
        current_time = self._player.playback_time
        self.play(self._video, current_time)
        return True

    @synchronous("_lock")
    def get_video_attr(self, attr: str, default=None):
        if self._video:
            return self._video.get_video_attr(attr, default)
        return default

    @synchronous("_lock")
    def configure_streams(self):
        audio_uid = self._video.aid
        sub_uid = self._video.sid

        if audio_uid is not None and not self._video.is_transcode:
            log.info("PlayerManager::play selecting audio stream index=%s" % audio_uid)
            self._player.audio = self._video.audio_seq[audio_uid]

        if sub_uid is None or sub_uid == -1:
            log.info("PlayerManager::play selecting subtitle stream (none)")
            self._player.sub = "no"
        else:
            log.info(
                "PlayerManager::play selecting subtitle stream index=%s" % sub_uid
            )
            if sub_uid in self._video.subtitle_seq:
                self._player.sub = self._video.subtitle_seq[sub_uid]
            elif sub_uid in self._video.subtitle_url:
                log.info(
                    "PlayerManager::play selecting external subtitle id=%s" % sub_uid
                )
                self.load_external_sub(sub_uid)

    @synchronous("_lock")
    def set_streams(self, audio_uid: int, sub_uid: int):
        need_restart = self._video.set_streams(audio_uid, sub_uid)

        if need_restart:
            self.restart_playback()
        else:
            self.configure_streams()
        self.timeline_handle()

    @synchronous("_lock")
    def load_external_sub(self, sub_id: int):
        if sub_id in self.external_subtitles:
            self._player.sub = self.external_subtitles[sub_id]
        else:
            try:
                sub_url = self._video.subtitle_url[sub_id]
                if settings.log_decisions:
                    log.info("Load External Subtitle: {0}".format(sub_url))
                self._player.sub_add(sub_url)
                self.external_subtitles[sub_id] = self._player.sub
                self.external_subtitles_rev[self._player.sub] = sub_id
            except SystemError:
                log.info("PlayerManager::could not load external subtitle")

    @synchronous("_lock")
    def toggle_fullscreen(self):
        self._player.fs = not self._player.fs
        self.fullscreen_disable = not self._player.fs

    @synchronous("_lock")
    def set_fullscreen(self, enabled: bool):
        self._player.fs = enabled
        self.fullscreen_disable = not enabled

    @synchronous("_lock")
    def set_mute(self, mute):
        self._player.mute = mute

    @synchronous("_lock")
    def screenshot(self):
        self._player.screenshot()

    @synchronous("_lock")
    def set_paused(self, value: bool, force: bool = False):
        if self.syncplay.is_enabled() and not force:
            if value:
                self.syncplay.pause_request()
            else:
                self.syncplay.play_request()
        else:
            self.pause_ignore = value
            self._player.pause = value

    @synchronous("_lock")
    def script_message(self, command, *args):
        self._player.command("script-message", command, *args)

    def get_track_ids(self):
        return self._video.aid, self._video.sid

    def update_subtitle_visuals(self):
        self._player.sub_pos = SUBTITLE_POS[settings.subtitle_position]
        self._player.sub_scale = settings.subtitle_size / 100
        self._player.sub_color = settings.subtitle_color
        self.timeline_handle()

    def get_timeline_options(self, finished=False):
        # PlaylistItemId is dynamically generated. A more stable Id will be used
        # if queue manipulation is added as a feature.
        player = self._player
        if finished:
            safe_pos = self._video.get_duration() or 0
        else:
            safe_pos = player.playback_time or 0
        self.last_seek = safe_pos
        self.pause_ignore = player.pause
        options = {
            "VolumeLevel": int(player.volume or 100),
            "IsMuted": player.mute,
            "IsPaused": player.pause,
            "RepeatMode": "RepeatNone",
            # "MaxStreamingBitrate": 140000000,
            "PositionTicks": int(safe_pos * 10000000),
            "PlaybackStartTimeTicks": int(self.start_time * 10000000),
            "SubtitleStreamIndex": none_fallback(self._video.sid, -1),
            "AudioStreamIndex": none_fallback(self._video.aid, -1),
            "BufferedRanges": [],
            "PlayMethod": "Transcode" if self._video.is_transcode else "DirectPlay",
            "PlaySessionId": self._video.playback_info["PlaySessionId"],
            "PlaylistItemId": self._video.get_playlist_id(),
            "MediaSourceId": self._video.media_source["Id"],
            "CanSeek": True,
            "ItemId": self._video.item_id,
            "NowPlayingQueue": self._video.parent.queue,
        }
        if player.duration is not None:
            options["BufferedRanges"] = [
                {
                    "start": int(safe_pos * 10000000),
                    "end": int(
                        (
                            (
                                player.duration
                                - safe_pos
                                * none_fallback(player.cache_buffering_state, 0)
                                / 100
                            )
                            + safe_pos
                        )
                        * 10000000
                    ),
                }
            ]
        if discord_presence:
            try:
                if (
                    self._video.is_tv
                    and self._video.item.get("IndexNumber") is not None
                    and self._video.item.get("ParentIndexNumber") is not None
                ):
                    title = self._video.item.get("SeriesName")
                    subtitle = _("Season {0} - Episode {1}").format(
                        self._video.item.get("ParentIndexNumber"),
                        self._video.item.get("IndexNumber"),
                    )
                else:
                    title = self._video.item.get("Name")
                    subtitle = str(self._video.item.get("ProductionYear", ""))
                send_presence(
                    title,
                    subtitle,
                    player.playback_time,
                    player.duration,
                    not player.pause,
                    self.syncplay.current_group,
                )
            except Exception:
                log.error("Could not send Discord Rich Presence.", exc_info=True)
        return options

    @synchronous("_tl_lock")
    def send_timeline(self):
        if (
            self.should_send_timeline
            and self._video
            and not self._player.playback_abort
        ):
            self._video.client.jellyfin.session_progress(self.get_timeline_options())
            try:
                if self.syncplay.is_enabled():
                    self.syncplay.sync_playback_time()
            except:
                log.error("Error syncing playback time.", exc_info=True)

    @synchronous("_tl_lock")
    def send_timeline_initial(self):
        self._video.client.jellyfin.session_playing(self.get_timeline_options())

    @synchronous("_tl_lock")
    def send_timeline_stopped(self, finished=False, options=None, client=None):
        self.should_send_timeline = False

        if options is None:
            options = self.get_timeline_options(finished)

        if client is None:
            client = self._video.client

        client.jellyfin.session_stop(options)

        if self.get_webview() is not None and settings.display_mirroring:
            self.get_webview().show()

        if discord_presence:
            try:
                clear_presence()
            except Exception:
                log.error("Could not clear Discord Rich Presence.", exc_info=True)

    def upd_player_hide(self):
        if self._video:
            self._player.keep_open = self._video.parent.has_next

    def terminate(self):
        self.stop()
        if is_using_ext_mpv:
            self._player.terminate()

        if self.trickplay:
            self.trickplay.stop()

    def get_seek_times(self):
        if self._jf_settings is None:
            self._jf_settings = self._video.client.jellyfin.get_user_settings()
        custom_prefs = self._jf_settings.get("CustomPrefs") or {}
        seek_left = custom_prefs.get("skipBackLength") or 15000
        seek_right = custom_prefs.get("skipForwardLength") or 30000
        return -int(seek_left) / 1000, int(seek_right) / 1000

    # Wrappers to avoid private access
    def is_active(self):
        return bool(self._player and self._video)

    def is_playing(self):
        return bool(self._video and not self._player.playback_abort)

    def is_not_paused(self):
        return bool(
            self._video and not self._player.playback_abort and not self._player.pause
        )

    def has_video(self):
        return self._video is not None

    def get_video(self):
        return self._video

    def show_text(self, text: str, duration: int, level: int = 1):
        self._player.show_text(text, duration, level)

    def get_osd_settings(self):
        return self._player.osd_back_color, self._player.osd_font_size

    def set_osd_settings(self, back_color: str, font_size: int):
        self._player.osd_back_color = back_color
        self._player.osd_font_size = font_size

    def enable_osc(self, enabled: bool):
        if settings.thumbnail_enable and self.trickplay:
            self.script_message(
                "osc-visibility", "auto" if enabled else "never", "False"
            )
        else:
            if hasattr(self._player, "osc"):
                self._player.osc = enabled

    def triggered_menu(self, enabled: bool):
        self.script_message("shim-menu-enable", "True" if enabled else "False")

    def playback_is_aborted(self):
        return self._player.playback_abort

    def force_window(self, enabled: bool):
        if enabled:
            self._player.force_window = True
            self._player.keep_open = True
            self._player.play("")
            if settings.fullscreen:
                self._player.fs = True
        else:
            self._player.keep_open = False
            if self._player.playback_abort:
                self._player.force_window = False
                self._player.play("")
            else:
                self.upd_player_hide()

    def add_ipc(self, ipc_name: str):
        self._player.input_ipc_server = ipc_name

    def get_current_client(self):
        return self._video.client

    def get_time(self):
        return self._player.playback_time

    def get_speed(self):
        return self._player.speed

    def set_speed(self, speed: float):
        self._player.speed = speed

    def kb_seek(self, action):
        if action == "up":
            self.seek(settings.seek_up, exact=settings.seek_v_exact)
        elif action == "down":
            self.seek(settings.seek_down, exact=settings.seek_v_exact)
        elif action == "left":
            seektime = settings.seek_left
            if settings.use_web_seek:
                seektime, _x = self.get_seek_times()
            self.seek(seektime, exact=settings.seek_h_exact)
        elif action == "right":
            seektime = settings.seek_right
            if settings.use_web_seek:
                _x, seektime = self.get_seek_times()
            self.seek(seektime, exact=settings.seek_h_exact)
        else:
            self.menu.menu_action(action)

    def menu_action(self, action):
        if self.menu.is_menu_shown:
            self.menu.menu_action(action)
        else:
            self.kb_seek(action)


playerManager = PlayerManager()



import logging
import os
import uuid
import socket
import json
import os.path
import sys
import getpass
from typing import Optional
from .settings_base import SettingsBase

log = logging.getLogger("conf")
config_path = None


def get_default_sdir():
    if sys.platform.startswith("win32"):
        if os.environ.get("USERPROFILE"):
            return os.path.join(os.environ["USERPROFILE"], "Desktop")
        else:
            username = getpass.getuser()
            return os.path.join(r"C:\Users", username, "Desktop")
    else:
        return None


class Settings(SettingsBase):
    player_name: str = socket.gethostname()
    audio_output: str = "hdmi"
    client_uuid: str = str(uuid.uuid4())
    media_ended_cmd: Optional[str] = None
    pre_media_cmd: Optional[str] = None
    stop_cmd: Optional[str] = None
    auto_play: bool = True
    idle_cmd: Optional[str] = None
    idle_ended_cmd: Optional[str] = None
    play_cmd: Optional[str] = None
    idle_cmd_delay: int = 60
    direct_paths: bool = False
    remote_direct_paths: bool = False
    always_transcode: bool = False
    transcode_hi10p: bool = False
    transcode_hdr: bool = False
    transcode_hevc: bool = False
    transcode_av1: bool = False
    transcode_4k: bool = False
    transcode_dolby_vision: bool = True
    allow_transcode_to_h265: bool = False
    prefer_transcode_to_h265: bool = False
    remote_kbps: int = 10000
    local_kbps: int = 2147483
    subtitle_size: int = 100
    subtitle_color: str = "#FFFFFFFF"
    subtitle_position: str = "bottom"
    fullscreen: bool = True
    enable_gui: bool = True
    media_key_seek: bool = False
    mpv_ext: bool = sys.platform.startswith("darwin")
    mpv_ext_path: Optional[str] = None
    mpv_ext_ipc: Optional[str] = None
    mpv_ext_start: bool = True
    mpv_ext_no_ovr: bool = False
    enable_osc: bool = True
    use_web_seek: bool = False
    display_mirroring: bool = False
    log_decisions: bool = False
    mpv_log_level: str = "info"
    idle_when_paused: bool = False
    stop_idle: bool = False
    kb_stop: str = "q"
    kb_prev: str = "<"
    kb_next: str = ">"
    kb_watched: str = "w"
    kb_unwatched: str = "u"
    kb_menu: str = "c"
    kb_menu_esc: str = "esc"
    kb_menu_ok: str = "enter"
    kb_menu_left: str = "left"
    kb_menu_right: str = "right"
    kb_menu_up: str = "up"
    kb_menu_down: str = "down"
    kb_pause: str = "space"
    kb_fullscreen: str = "f"
    kb_debug: str = "~"
    kb_kill_shader: str = "k"
    seek_up: int = 60
    seek_down: int = -60
    seek_right: int = 5
    seek_left: int = -5
    seek_v_exact: bool = False
    seek_h_exact: bool = False
    shader_pack_enable: bool = True
    shader_pack_custom: bool = False
    shader_pack_remember: bool = True
    shader_pack_profile: Optional[str] = None
    shader_pack_subtype: str = "lq"
    svp_enable: bool = False
    svp_url: str = "http://127.0.0.1:9901/"
    svp_socket: Optional[str] = None
    sanitize_output: bool = True
    write_logs: bool = False
    playback_timeout: int = 30
    sync_max_delay_speed: int = 50
    sync_max_delay_skip: int = 300
    sync_method_thresh: int = 2000
    sync_speed_time: int = 1000
    sync_speed_attempts: int = 3
    sync_attempts: int = 5
    sync_revert_seek: bool = True
    sync_osd_message: bool = True
    screenshot_menu: bool = True
    check_updates: bool = True
    notify_updates: bool = True
    lang: Optional[str] = None
    discord_presence: bool = False
    ignore_ssl_cert: bool = False
    menu_mouse: bool = True
    media_keys: bool = True
    connect_retry_mins: int = 0
    transcode_warning: bool = True
    lang_filter: str = "und,eng,jpn,mis,mul,zxx"
    lang_filter_sub: bool = False
    lang_filter_audio: bool = False
    force_set_played: bool = False
    screenshot_dir: Optional[str] = get_default_sdir()
    raise_mpv: bool = True
    force_video_codec: Optional[str] = None
    force_audio_codec: Optional[str] = None
    health_check_interval: Optional[int] = 300
    skip_intro_always: bool = False
    skip_intro_enable: bool = True
    skip_credits_always: bool = False
    skip_credits_enable: bool = True
    thumbnail_enable: bool = True
    thumbnail_osc_builtin: bool = True
    thumbnail_preferred_size: int = 320
    tls_client_cert: Optional[str] = None
    tls_client_key: Optional[str] = None
    tls_server_ca: Optional[str] = None

    def __get_file(self, path: str, mode: str = "r", create: bool = True):
        created = False

        if not os.path.exists(path):
            try:
                _fh = open(path, mode)
            except IOError as e:
                if e.errno == 2 and create:
                    fh = open(path, "w")
                    json.dump(self.dict(), fh, indent=4, sort_keys=True)
                    fh.close()
                    created = True
                else:
                    raise e
            except Exception:
                log.error("Error opening settings from path: %s" % path)
                return None

        # This should work now
        return open(path, mode), created

    def load(self, path: str, create: bool = True):
        global config_path  # Don't want in model.
        fh, created = self.__get_file(path, "r", create)
        config_path = path
        if not created:
            try:
                data = json.load(fh)
                safe_data = self.parse_obj(data)

                # Copy and count items
                input_params = 0
                for key in safe_data.__fields_set__:
                    setattr(self, key, getattr(safe_data, key))
                    input_params += 1

                # Print warnings
                for key, value in data.items():
                    if key not in safe_data.__fields_set__:
                        log.warning("Config item {0} was ignored.".format(key))
                    elif value != getattr(safe_data, key):
                        log.warning(
                            "Config item {0} was was coerced from {1} to {2}.".format(
                                key, repr(value), repr(getattr(safe_data, key))
                            )
                        )

                log.info("Loaded settings from json: %s" % path)
                if input_params < len(self.__fields__):
                    log.info("Saving back due to schema change.")
                    self.save()
            except Exception as e:
                log.error("Error loading settings from json: %s" % e)
                fh.close()
                return False

        fh.close()
        return True

    def save(self):
        if config_path is None:
            raise FileNotFoundError("Config path not set.")

        # noinspection PyTypeChecker
        fh, created = self.__get_file(config_path, "w", True)

        try:
            json.dump(self.dict(), fh, indent=4, sort_keys=True)
            fh.flush()
            fh.close()
        except Exception as e:
            log.error("Error saving settings to json: %s" % e)
            return False

        return True


settings = Settings()



import gettext
import locale

from .conf import settings

translation = gettext.NullTranslations()


def configure():
    global translation
    from .utils import get_resource

    messages_dir = get_resource("messages")
    lang = None

    if settings.lang is not None:
        lang = settings.lang
    else:
        # This is more robust than the built-in language detection in gettext.
        # Specifically, it supports Windows correctly.
        lc = locale.getdefaultlocale()
        if lc is not None and lc[0] is not None:
            lang = lc[0]

    if lang is not None:
        translation = gettext.translation(
            "base", messages_dir, languages=[lang], fallback=True
        )
    else:
        translation = gettext.translation("base", messages_dir, fallback=True)


def get_translation():
    return translation


def _(string: str) -> str:
    return translation.gettext(string)



import threading

import jinja2  # python3-jinja2 in Debian, Jinja2 in pypi

# So, most of my use of the webview library is ugly but there's a reason for this!
# Debian's python3-webview package is super old (2.3), pip3's pywebview package is much newer (3.2).
# I would just say "fuck it, install from testing/backports/unstable" except that's not any newer either.
# So instead I'm going to try supporting *both* here, which gets rather ugly because they made very significant changes.
#
# The key differences seem to be:
# 3.2's create_window() returns a Window object immediately, then you need to call start()
# 2.3's create_window() blocks forever effectivelly calling start() itself
#
# 3.2's Window has a .loaded Event that you need to subscribe to notice when the window is ready for input
# 2.3 has a webview_ready() function that blocks until webview is ready (or timeout is passed)
import webview  # Python3-webview in Debian, pywebview in pypi

from ..utils import get_text
from ..i18n import _
from . import helpers

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from jellyfin_apiclient_python import client as client_type

# This makes me rather uncomfortable, but there's no easy way around this other than
# importing display_mirror in helpers. Lambda needed because the 2.3 version of the JS
# api adds an argument even when not used.
helpers.on_escape = lambda _x=None: load_idle()


class DisplayMirror(object):
    display_window = None

    def __init__(self):
        self.open_player_menu = lambda: None
        self.webview = None

    def get_webview(self):
        return self.webview

    # noinspection PyUnresolvedReferences
    def run(self):
        # Webview needs to be run in the MainThread.

        # Prepare for version 2.3 before calling create_window(), which might block forever.
        self.display_window = webview
        # Since webview.create_window might take exclusive and permanent lock on the main thread,
        # we need to start this wait_load function before we start webview itself.
        if "webview_ready" in dir(webview):
            threading.Thread(
                target=lambda: (webview.webview_ready(), load_idle())
            ).start()

        window = webview.create_window(
            title="Jellyfin MPV Shim Mirror", js_api=helpers, fullscreen=True
        )
        if window is not None:
            # It returned a Window object instead of blocking, we're running on 3.2 (or compatible)
            self.display_window = window
            self.webview = window

            # 3.2's .loaded event runs every time a new DOM is loaded as well, so not suitable for this purpose
            # However, 3.2's load_html waits for the DOM to be ready, so we can completely skip
            # waiting for that ourselves.
            threading.Thread(target=load_idle).start()

            webview.start()

    def stop(self):
        if hasattr(webview, "destroy_window"):
            getattr(webview, "destroy_window")()
        else:
            self.webview.destroy()

    def display_content(self, client: "client_type", arguments):
        item = client.jellyfin.get_item(arguments["Arguments"]["ItemId"])
        html = get_html(server_address=client.config.data["auth.server"], item=item)
        self.display_window.load_html(html)
        # print(html)
        # breakpoint()
        return


mirror = DisplayMirror()


# FIXME: Add some support for some sort of theming beyond Jellyfin's css, to select user defined templates
def get_html(server_address: str = None, item=None):
    if item:
        jinja_vars = {
            "backdrop_src": helpers.getBackdropUrl(item, server_address) or "",
            "image_src": helpers.getPrimaryImageUrl(item, server_address) or "",
            "logo_src": helpers.getLogoUrl(item, server_address) or "",
            "played": item["UserData"].get("Played", False),
            "played_percentage": item["UserData"].get("PlayedPercentage", 0),
            "unplayed_items": item["UserData"].get("UnplayedItemCount", 0),
            "is_folder": item["IsFolder"],
            "display_name": helpers.getDisplayName(item),
            "misc_info_html": helpers.getMiscInfoHtml(item),
            "rating_html": helpers.getRatingHtml(item),
            "genres": item["Genres"],
            "overview": item.get("Overview", ""),
            # I believe these are all specifically for albums
            "poster_src": helpers.getPrimaryImageUrl(item, server_address) or "",
            "title": "title",  # FIXME
            "secondary_title": "secondary",  # FIXME
            "artist": "artist",  # FIXME
            "album_title": "album",  # FIXME
        }
    else:
        jinja_vars = {
            "random_backdrop": True,  # Make the jinja template load some extra JS code for random backdrops
            "backdrop_src": helpers.getRandomBackdropUrl(),  # Preinitialise it with a random backdrop though
            "display_name": _("Ready to cast"),
            "overview": "\n\n"
            + _(
                "Select your media in Jellyfin and play it here"
            ),  # FIME: Mention the player_name here
        }

    jinja_vars.update({"jellyfin_css": get_text("display_mirror", "jellyfin.css")})

    try:
        tpl = jinja2.Template(get_text("display_mirror", "index.html"))
        return tpl.render(jinja_vars)
    except Exception:
        pass


def load_idle():
    # FIXME: Add support for not actually having an idle screen and instead hide/close/something the window
    # Load the initial page before displaying any content,
    # and when refreshing to a blank page after idling.
    html = get_html()
    mirror.display_window.load_html(html)



import random
import datetime
import math

from ..clients import clientManager

# This started as a copy of some useful functions from jellyfin-chromecast's helpers.js and translated them to Python.
# Only reason their not put straight into __init__.py is to keep the same logical separation that
# jellyfin-chromecast has.
#
# I've since added some extra functions, and completely reworked some of the old ones such that it's
# not directly compatible.
#
# Should this stuff be in jellyfin_apiclient_python instead?
# Is this stuff in there already?
#
# FIXME: A lot of this could be done so much better with format-strings


# noinspection PyPep8Naming,PyPep8Naming
def getUrl(serverAddress, name):
    if not name:
        raise Exception("Url name cannot be empty")

    url = serverAddress
    url += "/" if not serverAddress.endswith("/") and not name.startswith("/") else ""
    url += name

    return url


# noinspection PyPep8Naming,PyPep8Naming
def getBackdropUrl(item, serverAddress):
    if item.get("BackdropImageTags"):
        return getUrl(
            serverAddress,
            "Items/"
            + item["Id"]
            + "/Images/Backdrop/0?tag="
            + item["BackdropImageTags"][0],
        )
    elif item.get("ParentBackdropItemId"):
        return getUrl(
            serverAddress,
            "Items/"
            + item["ParentBackdropItemId"]
            + "/Images/Backdrop/0?tag="
            + item["ParentBackdropImageTags"][0],
        )
    else:
        return None


# noinspection PyPep8Naming,PyPep8Naming
def getLogoUrl(item, serverAddress):
    if item.get("ImageTags", {}).get("Logo", None):
        return getUrl(
            serverAddress,
            "Items/" + item["Id"] + "/Images/Logo/0?tag=" + item["ImageTags"]["Logo"],
        )
    elif item.get("ParentLogoItemId") and item.get("ParentLogoImageTag"):
        return getUrl(
            serverAddress,
            "Items/"
            + item["ParentLogoItemId"]
            + "/Images/Logo/0?tag="
            + item["ParentLogoImageTag"],
        )
    else:
        return None


# noinspection PyPep8Naming,PyPep8Naming
def getPrimaryImageUrl(item, serverAddress):
    if item.get("AlbumPrimaryImageTag"):
        return getUrl(
            serverAddress,
            "Items/"
            + item["AlbumId"]
            + "/Images/Primary?tag="
            + item["AlbumPrimaryImageTag"],
        )
    elif item.get("PrimaryImageTag"):
        return getUrl(
            serverAddress,
            "Items/" + item["Id"] + "/Images/Primary?tag=" + item["PrimaryImageTag"],
        )
    elif item.get("ImageTags", {}).get("Primary"):
        return getUrl(
            serverAddress,
            "Items/"
            + item["Id"]
            + "/Images/Primary?tag="
            + item["ImageTags"]["Primary"],
        )
    else:
        return None


# noinspection PyPep8Naming
def getDisplayName(item):
    name = item.get("EpisodeTitle", item.get("Name"))

    if item["Type"] == "TvChannel":
        if item["Number"]:
            return item["Number"] + " " + name
        else:
            return name
    # NOTE: Must compare to None here because 0 is a legitimate option
    elif (
        item["Type"] == "Episode"
        and item.get("IndexNumber") is not None
        and item.get("ParentIndexNumber") is not None
    ):
        number = f"S{item['ParentIndexNumber']} E{item['IndexNumber']}"
        if item.get("IndexNumberEnd"):
            number += "-" + item["IndexNumberEnd"]
        name = number + " - " + name

    return name


# noinspection PyPep8Naming
def getRatingHtml(item):
    html = ""

    if item.get("CommunityRating"):
        html += (
            "<div class='starRating' title='"
            + str(item["CommunityRating"])
            + "'></div>"
        )
        html += '<div class="starRatingValue">'
        html += str(round(item["CommunityRating"], 1))
        html += "</div>"

    if item.get("CriticRating") is not None:
        # FIXME: This doesn't seem to ever be triggering. Is that actually a problem?
        if item["CriticRating"] >= 60:
            html += '<div class="fresh rottentomatoesicon" title="fresh"></div>'
        else:
            html += '<div class="rotten rottentomatoesicon" title="rotten"></div>'

        html += '<div class="criticRating">' + str(item["CriticRating"]) + "%</div>"

    # Jellyfin-chromecast had this commented out already
    # Where's the metascore variable supposed to come from?
    # # if item.get(Metascore) and metascore !== false: {
    # #     if item['Metascore'] >= 60:
    # #         html += '<div class="metascore metascorehigh" title="Metascore">' + item['Metascore'] + '</div>'
    # #     elif item['Metascore'] >= 40):
    # #         html += '<div class="metascore metascoremid"  title="Metascore">' + item['Metascore'] + '</div>'
    # #     else:
    # #         html += '<div class="metascore metascorelow"  title="Metascore">' + item['Metascore'] + '</div>'

    return html


def __convert_jf_str_datetime(jf_string):
    # datetime doesn't quite support fractions of a second the same way Jellyfin does them.
    # Best we can do is strip them out entirely.
    # FIXME: I think this loses timezone information, but are we getting any at all anyway?
    return datetime.datetime.strptime(jf_string.partition(".")[0], "%Y-%m-%dT%H:%M:%S")


# noinspection PyPep8Naming,PyPep8Naming,PyPep8Naming
def getMiscInfoHtml(item):
    # FIXME: Flake8 is complaining this function is too complex.
    #        I agree, this needs to be cleaned up, a lot.
    # FIXME: This shouldn't return HTML, the template should take care of that.

    miscInfo = []

    if item["Type"] == "Episode":
        if item.get("PremiereDate"):
            date = __convert_jf_str_datetime(item["PremiereDate"])
            text = date.strftime("%x")
            miscInfo.append(text)

    if item.get("StartDate"):
        date = __convert_jf_str_datetime(item["StartDate"])
        text = date.strftime("%x")
        miscInfo.append(text)

        if item["Type"] != "Recording":
            pass
            # Jellyfin-chromecast had this commented out already
            # # text = LiveTvHelpers.getDisplayTime(date)
            # # miscInfo.push(text)

    if item.get("ProductionYear") and item["Type"] == "Series":
        if item["Status"] == "Continuing":
            miscInfo.append(f"{item['ProductionYear']}-Present")
        elif item["ProductionYear"]:
            text = str(item["ProductionYear"])
            if item.get("EndDate"):
                endYear = __convert_jf_str_datetime(item["EndDate"]).year
                if endYear != item["ProductionYear"]:
                    text += "-" + str(endYear)
            miscInfo.append(text)

    if item["Type"] != "Series" and item["Type"] != "Episode":
        if item.get("ProductionYear"):
            miscInfo.append(str(item["ProductionYear"]))
        elif item.get("PremiereDate"):
            text = str(__convert_jf_str_datetime(item["PremiereDate"]).year)
            miscInfo.append(text)
    if item.get("RunTimeTicks") and item["Type"] != "Series":
        if item["Type"] == "Audio":
            # FIXME
            raise Exception("Haven't translated this to Python yet")
            # miscInfo.append(datetime.getDisplayRunningTime(item["RunTimeTicks"]))
        else:
            # Using math.ceil instead of round because I want the minutes rounded *up* specifically,
            # mostly because '1min' makes more sense than '0min' for a 1-59sec clip
            # FIXME: Alternatively, display '<1min' if it's 0?
            minutes = math.ceil(item["RunTimeTicks"] / 600000000)
            miscInfo.append(f"{minutes}min")

    if (
        item.get("OfficialRating")
        and item["Type"] != "Season"
        and item["Type"] != "Episode"
    ):
        miscInfo.append(item["OfficialRating"])

    if item.get("Video3DFormat"):
        miscInfo.append("3D")

    return "&nbsp;&nbsp;&nbsp;&nbsp;".join(miscInfo)


# For some reason the webview 2.3 js api will send a positional argument of None when there are no
# arguments being passed in. This really long argument name is here to catch that and hopefully not
# eat other intentional arguments.
# noinspection PyPep8Naming
def getRandomBackdropUrl(_positional_arg_that_is_never_used=None, **params):
    # This function is to get 1 random item, so ignore those arguments
    params["SortBy"] = "Random"
    params["Limit"] = 1

    # Use sensible defaults for all other arguments.
    # Based on jellyfin-chromecast's behaviour.
    params["IncludeItemTypes"] = params.get("IncludeItemTypes", "Movie,Series")
    params["ImageTypes"] = params.get("ImageTypes", "Backdrop")
    params["Recursive"] = params.get("Recursive", True)
    params["MaxOfficialRating"] = params.get("MaxOfficialRating", "PG-13")

    # This application can have multiple client connections different servers at the same time.
    # So just pick a random one of those clients to query for the random item.
    client = random.choice(list(clientManager.clients.values()))
    item = client.jellyfin.user_items(params=params)["Items"][0]

    return getBackdropUrl(item, client.config.data["auth.server"])



APP_NAME = "jellyfin-mpv-shim"
USER_APP_NAME = "Jellyfin MPV Shim"
CLIENT_VERSION = "2.9.0"
USER_AGENT = "Jellyfin-MPV-Shim/%s" % CLIENT_VERSION
CAPABILITIES = {
    "PlayableMediaTypes": "Video",
    "SupportsMediaControl": True,
    "SupportedCommands": (
        "MoveUp,MoveDown,MoveLeft,MoveRight,Select,"
        "Back,ToggleFullscreen,"
        "GoHome,GoToSettings,TakeScreenshot,"
        "VolumeUp,VolumeDown,ToggleMute,"
        "SetAudioStreamIndex,SetSubtitleStreamIndex,"
        "Mute,Unmute,SetVolume,DisplayContent,"
        "Play,Playstate,PlayNext,PlayMediaSource"
    ),
}



import logging
import urllib.parse
import os.path
import re
import pathlib
from io import BytesIO
from sys import platform

from .conf import settings
from .utils import is_local_domain, get_profile, get_seq
from .i18n import _

log = logging.getLogger("media")

from typing import TYPE_CHECKING, Optional, List

if TYPE_CHECKING:
    from jellyfin_apiclient_python import JellyfinClient as JellyfinClient_type


class Intro(object):
    def __init__(self, type, start, end):
        self.type: str = type  # "Intro" or "Outro"
        self.start: float = start
        self.end: float = end
        self.has_triggered: bool = False


class Video(object):
    def __init__(
        self,
        item_id: str,
        parent: "Media",
        aid: Optional[int] = None,
        sid: Optional[int] = None,
        srcid: Optional[str] = None,
    ):
        self.item_id = item_id
        self.parent = parent
        self.client = parent.client
        self.aid = aid
        self.sid = sid
        self.item = self.client.jellyfin.get_item(item_id)

        self.is_tv = self.item.get("Type") == "Episode"

        self.subtitle_seq = {}
        self.subtitle_uid = {}
        self.subtitle_url = {}
        self.subtitle_enc = set()
        self.audio_seq = {}
        self.audio_uid = {}
        self.is_transcode = False
        self.trs_ovr = None
        self.playback_info = None
        self.media_source = None
        self.srcid = srcid
        self.intros: List[Intro] = []
        self.intro_tried = False

    def map_streams(self):
        self.subtitle_seq = {}
        self.subtitle_uid = {}
        self.subtitle_url = {}
        self.subtitle_enc = set()
        self.audio_seq = {}
        self.audio_uid = {}

        if self.media_source is None or self.media_source["Protocol"] != "File":
            return

        index = 1
        for stream in self.media_source["MediaStreams"]:
            if stream.get("Type") != "Audio":
                continue

            self.audio_uid[index] = stream["Index"]
            self.audio_seq[stream["Index"]] = index

            if not stream.get("IsExternal"):
                index += 1

        index = 1
        for sub in self.media_source["MediaStreams"]:
            if sub.get("Type") != "Subtitle":
                continue

            if sub.get("DeliveryMethod") == "Embed":
                self.subtitle_uid[index] = sub["Index"]
                self.subtitle_seq[sub["Index"]] = index
            elif sub.get("DeliveryMethod") == "External":
                url = sub.get("DeliveryUrl")
                if not sub.get("IsExternalUrl"):
                    url = self.client.config.data["auth.server"] + url
                self.subtitle_url[sub["Index"]] = url
            elif sub.get("DeliveryMethod") == "Encode":
                self.subtitle_enc.add(sub["Index"])

            if not sub.get("IsExternal"):
                index += 1

        user_aid = self.media_source.get("DefaultAudioStreamIndex")
        user_sid = self.media_source.get("DefaultSubtitleStreamIndex")

        if user_aid is not None and self.aid is None:
            self.aid = user_aid

        if user_sid is not None and self.sid is None:
            self.sid = user_sid

    def get_current_streams(self):
        return self.aid, self.sid

    def get_proper_title(self):
        if not hasattr(self, "_title"):
            title = self.item.get("Name")
            if (
                self.is_tv
                and self.item.get("IndexNumber") is not None
                and self.item.get("ParentIndexNumber") is not None
            ):
                episode_number = int(self.item.get("IndexNumber"))
                season_number = int(self.item.get("ParentIndexNumber"))
                series_name = self.item.get("SeriesName")
                title = "%s - s%de%.2d - %s" % (
                    series_name,
                    season_number,
                    episode_number,
                    title,
                )
            elif self.item.get("Type") == "Movie":
                year = self.item.get("ProductionYear")
                if year is not None:
                    title = "%s (%s)" % (title, year)
            setattr(self, "_title", title)
        return getattr(self, "_title") + (
            _(" (Transcode)") if self.is_transcode else ""
        )

    def set_trs_override(self, video_bitrate: Optional[int], force_transcode: bool):
        if force_transcode:
            self.trs_ovr = (video_bitrate, force_transcode)
        else:
            self.trs_ovr = None

    def get_transcode_bitrate(self):
        if not self.is_transcode:
            return "none"
        elif self.trs_ovr is not None:
            if self.trs_ovr[0] is not None:
                return self.trs_ovr[0]
            elif self.trs_ovr[1]:
                return "max"
        elif self.parent.is_local:
            return "max"
        else:
            return settings.remote_kbps

    def terminate_transcode(self):
        if self.is_transcode:
            try:
                self.client.jellyfin.close_transcode(
                    self.client.config.data["app.device_id"]
                )
            except:
                log.warning("Terminating transcode failed.", exc_info=1)

    def _get_url_from_source(self):
        # Only use Direct Paths if:
        # - The media source supports direct paths.
        # - Direct paths are enabled in the config.
        # - The server is local or the override config is set.
        # - If there's a scheme specified or the path exists as a local file.
        if (
            (
                self.media_source.get("Protocol") == "Http"
                or self.media_source["SupportsDirectPlay"]
            )
            and settings.direct_paths
            and (settings.remote_direct_paths or self.parent.is_local)
        ):
            if platform.startswith("win32") or platform.startswith("cygwin"):
                # matches on SMB scheme
                match = re.search("(?:\\\\).+:.*@(.+)", self.media_source["Path"])
                if match:
                    # replace forward slash to backward slashes
                    log.debug("cleaned up credentials from path")
                    self.media_source["Path"] = str(
                        pathlib.Path("\\\\" + match.groups()[0])
                    )

            if urllib.parse.urlparse(self.media_source["Path"]).scheme:
                self.is_transcode = False
                log.debug("Using remote direct path.")
                # translate path for windows
                # if path is smb path in credential format for kodi and maybe linux \\username:password@mediaserver\foo,
                # translate it to mediaserver/foo
                return str(pathlib.Path(self.media_source["Path"]))
            else:
                # If there's no uri scheme, check if the file exixsts because it might not be mounted
                if os.path.isfile(self.media_source["Path"]):
                    log.debug("Using local direct path.")
                    self.is_transcode = False
                    return self.media_source["Path"]

        if self.media_source["SupportsDirectStream"]:
            self.is_transcode = False
            log.info("Using direct url.")
            query_params = {
                "static": "true",
                "MediaSourceId": self.media_source["Id"],
                "api_key": self.client.config.data["auth.token"],
            }

            if "LiveStreamId" in self.media_source:
                query_params["LiveStreamId"] = self.media_source["LiveStreamId"]

            query = urllib.parse.urlencode(query_params)

            return "%s/Videos/%s/stream?%s" % (
                self.client.config.data["auth.server"],
                self.item_id,
                query,
            )
        elif self.media_source["SupportsTranscoding"]:
            log.info("Using transcode url.")
            self.is_transcode = True
            return self.client.config.data["auth.server"] + self.media_source.get(
                "TranscodingUrl"
            )

    def get_best_media_source(self, preferred: Optional[str] = None):
        weight_selected = 0
        preferred_selected = None
        selected = None
        for media_source in self.playback_info["MediaSources"]:
            if media_source.get("Id") == preferred:
                preferred_selected = media_source
            # Prefer the highest bitrate file that will direct play.
            weight = (media_source.get("SupportsDirectPlay") or 0) * 50000 + (
                media_source.get("Bitrate") or 0
            ) / 1000
            if weight > weight_selected:
                weight_selected = weight
                selected = media_source
        if preferred_selected:
            return preferred_selected
        else:
            if preferred is not None:
                log.warning("Preferred media source is unplayable.")
            return selected

    def get_intro(self, media_source_id):
        if self.intro_tried:
            return
        self.intro_tried = True

        # provided by plugin
        try:
            skip_intro_data = self.client.jellyfin.media_segments(
                f"/{media_source_id}?includeSegmentTypes=Outro&includeSegmentTypes=Intro"
            )
            for intro in skip_intro_data["Items"]:
                self.intros.append(
                    Intro(
                        intro["Type"], # Intro or Outro
                        intro["StartTicks"] / 10000000,
                        intro["EndTicks"] / 10000000
                    )
                )
        except:
            log.warning(
                "Fetching intro data failed.",
                exc_info=1,
            )

    def get_current_intro(self, time):
        for intro in self.intros:
            if intro.start <= time and time <= intro.end:
                return intro.start <= time, intro
        return False, None

    def get_chapters(self):
        return [
            {"start": item["StartPositionTicks"] / 10000000, "name": item["Name"]}
            for item in self.item.get("Chapters", [])
            if item.get("ImageTag")
        ]

    def get_chapter_images(self, max_width=400, quality=90):
        for i, item in enumerate(self.item.get("Chapters", [])):
            data = BytesIO()
            self.client.jellyfin._get_stream(
                f"Items/{self.item_id}/Images/Chapter/{i}",
                data,
                {"tag": item["ImageTag"], "maxWidth": max_width, "quality": quality},
            )
            yield data.getvalue()

    def get_hls_tile_images(self, width, count):
        for i in range(0, count):
            data = BytesIO()
            self.client.jellyfin._get_stream(
                f"Videos/{self.item['Id']}/Trickplay/{width}/{i}.jpg?MediaSourceId={self.media_source['Id']}",
                data,
            )
            yield data.getvalue()

    def get_bif(self, prefer_width=320):
        manifest = self.item.get("Trickplay")
        if (
            manifest is not None
            and manifest.get(self.media_source["Id"]) is not None
            and len(manifest[self.media_source["Id"]]) > 0
        ):
            available_widths = [
                int(x) for x in manifest[self.media_source["Id"]].keys()
            ]

            if prefer_width is not None:
                width = min(available_widths, key=lambda x: abs(x - prefer_width))
            else:
                width = max(available_widths)

            return manifest[self.media_source["Id"]][str(width)]
        else:
            return None

    def get_playback_url(
        self,
        video_bitrate: Optional[int] = None,
        force_transcode: Optional[int] = False,
    ):
        """
        Returns the URL to use for the transcoded file.
        """
        self.terminate_transcode()

        if self.trs_ovr:
            video_bitrate, force_transcode = self.trs_ovr

        log.info(
            "Bandwidth: local={0}, bitrate={1}, force={2}".format(
                self.parent.is_local, video_bitrate, force_transcode
            )
        )
        profile = get_profile(not self.parent.is_local, video_bitrate, force_transcode)
        self.playback_info = self.client.jellyfin.get_play_info(
            self.item_id, profile, self.aid, self.sid, media_source_id=self.srcid
        )

        self.media_source = self.get_best_media_source(self.srcid)
        if (
            settings.skip_intro_always
            or settings.skip_intro_enable
            or settings.skip_credits_always
            or settings.skip_credits_enable
        ):
            self.get_intro(self.media_source["Id"])

        self.map_streams()
        url = self._get_url_from_source()

        # If there are more media sources and the default one fails, try all of them.
        if url is None and len(self.playback_info["MediaSources"]) > 1:
            log.warning("Selected media source is unplayable.")
            for media_source in self.playback_info["MediaSources"]:
                if media_source["Id"] != self.srcid:
                    self.media_source = media_source
                    self.map_streams()
                    url = self._get_url_from_source()
                    if url is not None:
                        break

        if settings.log_decisions:
            if len(self.playback_info["MediaSources"]) > 1:
                log.info("Full Playback Info: {0}".format(self.playback_info))
            log.info("Media Decision: {0}".format(self.media_source))
        return url

    def get_duration(self):
        ticks = self.item.get("RunTimeTicks")
        if ticks:
            return ticks / 10000000

    def set_played(self, watched: bool = True):
        self.client.jellyfin.item_played(self.item_id, watched)

    def set_streams(self, aid: Optional[int], sid: Optional[int]):
        need_restart = False

        if aid is not None and self.aid != aid:
            self.aid = aid
            if self.is_transcode:
                need_restart = True

        if sid is not None and self.sid != sid:
            self.sid = sid
            if sid in self.subtitle_enc:
                need_restart = True

        return need_restart

    def get_playlist_id(self):
        return self.parent.queue[self.parent.seq]["PlaylistItemId"]


class Media(object):
    def __init__(
        self,
        client: "JellyfinClient_type",
        queue: list,
        seq: int = 0,
        user_id: Optional[str] = None,
        aid: Optional[int] = None,
        sid: Optional[int] = None,
        srcid: Optional[str] = None,
        queue_override: bool = True,
    ):
        if queue_override:
            self.queue = [
                {"PlaylistItemId": "playlistItem{0}".format(get_seq()), "Id": id_num}
                for id_num in queue
            ]
        else:
            self.queue = queue
        self.client = client
        self.seq = seq
        self.user_id = user_id

        self.video = Video(self.queue[seq]["Id"], self, aid, sid, srcid)
        self.is_tv = self.video.is_tv
        self.is_local = is_local_domain(client)
        self.has_next = seq < len(queue) - 1
        self.has_prev = seq > 0

    def get_next(self):
        if self.has_next:
            return Media(
                self.client,
                self.queue,
                self.seq + 1,
                self.user_id,
                queue_override=False,
            )

    def get_prev(self):
        if self.has_prev:
            return Media(
                self.client,
                self.queue,
                self.seq - 1,
                self.user_id,
                queue_override=False,
            )

    def get_from_key(self, item_id: str):
        for i, video in enumerate(self.queue):
            if video["Id"] == item_id:
                return Media(
                    self.client, self.queue, i, self.user_id, queue_override=False
                )
        return None

    def get_video(self, index: int):
        if index == 0 and self.video:
            return self.video

        if index < len(self.queue):
            return Video(self.queue[index]["Id"], self)

        log.error("Media::get_video couldn't find video at index %s" % index)

    def insert_items(self, items, append: bool = False):
        items = [
            {"PlaylistItemId": "playlistItem{0}".format(get_seq()), "Id": id_num}
            for id_num in items
        ]
        if append:
            self.queue.extend(items)
        else:
            self.queue = (
                self.queue[0 : self.seq + 1] + items + self.queue[self.seq + 1 :]
            )
        self.has_next = self.seq < len(self.queue) - 1

    def replace_queue(self, sp_items, seq):
        """Update queue for SyncPlay.
        Returns None if the video is the same or a new Media if not."""
        if self.queue[self.seq]["Id"] == sp_items[seq]["Id"]:
            self.queue, self.seq = sp_items, seq
            return None
        else:
            return Media(self.client, sp_items, seq, self.user_id, queue_override=False)
